;compiler: avrasm
	.include "m644def.inc"
	.include "macros.asm"
	
	;technical settings
	.equ	XTAL=20000000	;XTAL frequency [Hz]
	.equ	HCOC=XTAL>24000000	;SRAM wait cycle if hardcore overclock (usualy if XTAL>24MHz) (1/0 - on/off)
	.equ	USRDIV=0	;USART baudrate divisor (baud_rate=XTAL/(16*(USRDIV+1)))
	.equ	MEMTST=0	;full range ext. SRAM test (1-yes, 0-no)
	.equ	STORA=1	;enable storage (1-yes, 0-no)
		
	;MOD player settings
	.equ	DAC=1545	;1543 for I2S or 1545 for TDA1545 DAC
	.equ	SRATE=21000	;NEEDED sample rate [Hz]
	.equ	MAXCH=8	;maximum valid count of channels (4-32)
	.equ	MAMPLD=MAXCH<=4	;channel amplitude (1/0 - full/half), full amplitude should be set only for 4 channel MODs  
	.equ	STEREO=1	;stereo/mono mixing (1/0 - stereo/mono)
	.equ	LINRES=1	;linear interpolation (1/0 - on/off, a 'bit' more CPU power needed)
	.equ	DRAWVU=1	;draw VU meters !!! +5% CPU usage (detecting peeks in resampling loop) !!! (1/0 - on/off)
			
	;other settings
	.equ	LING=0	;text language (0 - english, 1 - cesky)
	


;************************************************************************************************
;	Signal definitions
;************************************************************************************************	
	;IDE signals
	.equ	IORD=3	;IDE read strobe
	.equ	pIORD=portd
	.equ	IOWR=7	;IDE write strobe
	.equ	pIOWR=portd
	.equ	IOA0=5	;IDE A0 on porta (physicaly conencted at high address latch)
	.equ	IOA1=6	;IDE A1 on porta (physicaly conencted at high address latch)
	.equ	IOA2=7	;IDE A2 on porta (physicaly conencted at high address latch)
		
	;LCD signals
	.equ	LRS=1	;LCD register selection (1/0 - data/instructions)
	.equ	pLRS=portb
	.equ	LCK=0	;LCD clock (positive pulse)
	.equ	pLCK=portb
	
	;DAC signals
	.equ	DWS=2	;DAC channel multiplex 
	.equ	pDWS=portb
	.equ	DCK=7	;DAC clock 
	.equ	pDCK=portb
	.equ	DDT=5	;DAC data 
	.equ	pDDT=portb
	.equ	DSS=4	;SPI SS (must be configured as output!)
	.equ	pDSS=portb
	
	;SRAM signals
	.equ	RWR=3	;SRAM write (negative pulse)
	.equ	pRWR=portb
	.equ	ROE=4	;SRAM output enable (negative active)
	.equ	pROE=portd
	.equ	RMA=6	;SRAM medium address latch clock (rissing edge sensitive)
	.equ	pRMA=portd
	.equ	RHA=5	;SRAM high address latch clock (rissing edge sensitive)
	.equ	pRHA=portd
	
	;buttons
	.equ	BCO=2	;Buttons common signal (active low)
	.equ	pBCO=portd
	.equ	dBCO=ddrd

	
	
	;RAM SEGMENT
	.dseg
	.org	0x0100	;user RAM
	.equ	STSIZE=256	;reserved stack memory (cause error if free memory less than STSIZE)
BtnInh:	.byte	1	;Button inhibitor
BtnDiv:	.byte	1	;Button irq divider
BtnLast:	.byte	1	;Last buttons state
BtnAutoS:	.byte	1	;buttons auto repeater step
BtnAutoD:	.byte	1	;buttons auto repeater delay counter
BtnAutoR:	.byte	1	;buttons auto repeater rate variable
SRAMsize:	.byte	1	;external SRAM size in 64kiB blocks
MODflag:	.byte	1	;MOD flags
	;0 - leave playback 
MODname:	.byte	20	;MOD name
MODinsn:	.byte	1	;MOD instrument count
MODchnn:	.byte	1	;MOD channel count
MODpatl:	.byte	128	;MOD pattern list
MODpatn:	.byte	1	;MOD pattern count (song len in patterns)
MODpats:	.byte	1	;MOD pattern count (real pattern count)
	.equ	MODsmplSize=1+3+3+3+1
MODsmpl:	.byte	MODsmplSize*31	;MOD sample info for max channel count
	;vol [0-64], begin [Bytes], end [Bytes], repeat [Bytes], repeat [1-on, 0-off]
MODspd:	.byte	1	;MOD speed (ticks per note cycle)
MODbpm:	.byte	1	;MOD beats per minute
MODtdiv:	.byte	2	;MOD tick divisor (2.5*sampling_rate/bpm)
MODtimc:	.byte	1	;MOD timing changed flag (1 - speed, 2 - tempo)
MODpat:	.byte	1	;MOD actual pattern (position in MODpatl)
MODpln:	.byte	1	;MOD actual pattern line
MODpnln:	.byte	1	;MOD new line for pattern break effect
MODpatp:	.byte	3	;MOD actual note pointer (SRAM)
MODncnt:	.byte	1	;MOD note rate divisor counter

	.equ	MODevolsSize=1+1
	.equ	MODeportSize=1+1
	.equ	MODearpSize=1+1+2+2
	.equ	MODevibSize=1+1+2+1+2
	.equ	MODeretSize=1+1+1+2
	.equ	MODeposSize=1	
	;MOD playback structure for each channel
	.equ	MODmplSize=1+1+2+4+3+3+1+1+2+2
	.equ	MODpbckSize=MODmplSize+(MODevolsSize+MODeportSize+MODearpSize+MODevibSize+MODeretSize+MODeposSize)
MODpbck:	.byte	MODpbckSize*MAXCH	
	;activity, vol, resample, pos, end, loop len, loop activity, last instrument, period, tone-port. final period, (effect data)
	.equ	mePB=0	;main playback data offset
	.equ	meVS=mePB+MODmplSize	;volume slide offset
	.equ	meTP=meVS+MODevolsSize	;tone portamento offset
	.equ	meAR=meTP+MODeportSize	;arpeggio offset
	.equ	meVI=meAR+MODearpSize	;vibrato offset
	.equ	meRT=meVI+MODevibSize	;retrigger offset
	.equ	mePS=meRT+MODeretSize	;sample position offset

	.if	DRAWVU==1
MODbars:	.byte	MAXCH	;MOD VU-meter peaks for each channel
	.endif
FileLen:	.byte	3	;MOD file len
TimOut:	.byte	1	;timer1 timeout flag
TimOutDiv:	.byte	1	;timeout timer divisor 
FIFOtwr:	.byte	1	;FIFO free to write [Byte]
FIFOrd:	.byte	2	;FIFO read pointer
FIFOwr:	.byte	2	;FIFO write pointer
	.equ	FIFOsize=2300	;FIFO size (multiple of 4!)
FIFO:	.byte	FIFOsize	;FIFO buffer
	.equ	DBuff=FIFO	;data buffer (overlapped with FIFO, size at least 1024B)


	;CODE SEGMENT
	.cseg
	.org	0x0000	;reset
	rjmp	Init
	.org	0x0012
	rjmp	Timer2ctc
	.org	0x001A
	rjmp	Timer1ctc
	.org	0x0020	;timer 0 ctc a
	rjmp	Timer0ctc
	

;************************************************************************************************
;	FIFO output
;************************************************************************************************
	;some realy usefull constants
	.if	XTAL/(8*SRATE)-1>255	;ocr constant higher than 255?
	.equ	SRDIV=64	;yeah - set timer 0 divisor /64
	.else
	.equ	SRDIV=8	;no - set timer 0 divisor /8
	.endif
	.equ	SROCR=XTAL/(SRDIV*SRATE)-1	;timer 0 ocr constant for SRATE sampling freq. and SRDIV divisor 
	.equ	SREAL=XTAL/(SRDIV*(SROCR+1))	;REAL sampling rate for SROCR ocr constant [Hz]
	.equ	PALF=256*3546895/SREAL	;magic constant for resampling coeff. calcualtion (PAL/2*real_sample_rate [fx ?.8])
	.equ	MINT=30	;minimal allowed period (usually 54)
	.equ	MAXT=4095	;maximum allowed period (usually 1814)
	
	
	;period to resamplimg coeff calculation R16:R17=PALF/T, where T=R16:R17
	.macro	T2RC
	.if	PALF<=65535
	;for PALF [fx 8.8]
	ldi	R16,low(PALF)
	ldi	R17,high(PALF)
	rcall	div16u
	.else
	;for PALF [fx 16.8]
	push	R13
	push	R20
	push	R21
	mov	R20,R19
	mov	R19,R18
	ldi	R21,0
	ldi	R16,low(PALF)
	ldi	R17,high(PALF)
	ldi	R18,byte3(PALF)
	rcall	div24u
	pop	R21
	pop	R20
	pop	R13
	.endif	
	.endm

	
	;irq DAC output routine (timer0 ocra interrupt)
Timer0ctc:	.if	DAC==1543
	;---------;
	; TDA1543 ;
	;---------;
	out	ocr0b,R16	;push
	in	R16,sreg
	out	gpior0,R16	;push
	out	gpior1,R30	;push
	out	gpior2,R31	;push
	;
	ldsw	R30,FIFOrd	;FIFO read pointer
	;
	ld	R16,Z+
	out	spdr,R16	;high byte left
	;15 cycles pause
	ldi	R16,5
	dbnz	R16,PC-1
	;
	ld	R16,Z+
	out	spdr,R16	;low byte left
	;12 cycles pause
	ldi	R16,3
	dbnz	R16,PC-1
	cpiwh	R30,R16,FIFO+FIFOsize-2
	;
	sbi	pDWS,DWS	;negate WS
	nop
	ld	R16,Z+
	out	spdr,R16	;high byte right
	ld	R16,Z+
	;15 cycles pause
	brne	PC+1+2+1
	ldiw	R30,FIFO
	rjmp	PC+1+1+1
	adiw	R30,0
	nop
	stsw	FIFOrd,R30	
	ldi	R16,2
	dbnz	R16,PC-1	;6 cycles
	;
	out	spdr,R16	;low byte right
	;15 cycles pause
	ldi	R16,2
	dbnz	R16,PC-1	;6 cycles
	nop
	in	R31,gpior2	;pop
	in	R30,gpior1	;pop
	in	R16,gpior0	;pop
	out	sreg,R16
	in	R16,ocr0b	;pop
	;
	cbi	pDWS,DWS	;negate WS
	

	.elif	DAC==1545
	;---------;
	; TDA1545 ;
	;---------;
	;irq DAC output routine (timer0 ocra interrupt)
	out	ocr0b,R16	;push
	in	R16,sreg
	out	gpior0,R16	;push
	out	gpior1,R30	;push
	out	gpior2,R31	;push
	;
	ldsw	R30,FIFOrd	;FIFO read pointer
	;
	ld	R16,Z+
	out	spdr,R16	;high byte left
	;15 cycles pause
	ldi	R16,5
	dbnz	R16,PC-1
	;
	ld	R16,Z+
	out	spdr,R16	;low byte left
	;15 cycles pause
	ldi	R16,4
	dbnz	R16,PC-1
	cpiwh	R30,R16,FIFO+FIFOsize-2
	;
	ld	R16,Z+
	sbi	pDWS,DWS	;negate WS
	out	spdr,R16	;high byte right
	ld	R16,Z+
	;15 cycles pause
	brne	PC+1+2+1
	ldiw	R30,FIFO
	rjmp	PC+1+1+1
	adiw	R30,0
	nop
	stsw	FIFOrd,R30
	ldi	R30,2
	dbnz	R30,PC-1	;6 cycles
	;
	out	spdr,R16	;low byte right
	;15 cycles pause
	ldi	R30,4
	dbnz	R30,PC-1
	nop
	in	R31,gpior2	;pop
	in	R30,gpior1	;pop
	in	R16,gpior0	;pop
	out	sreg,R16
	in	R16,ocr0b	;pop
	;
	cbi	pDWS,DWS	;negate WS
	
	.else
	.error	"INVALID DAC SELECTION"
	.endif
	reti
	

	;start DAC output
	.macro	DArun
	lds	R16,timsk0
	sbr	R16,1<<ocie0a
	sts	timsk0,R16
	.endm
	
	;stop DAC output
	.macro	DAstop
	lds	R16,timsk0
	cbr	R16,1<<ocie0a
	sts	timsk0,R16
	.endm

	;FIFO clear
FIFOclr:	ldiw	R30,FIFO
	ldiw	R24,FIFOsize
	ldi	R16,0
FIFOclrL:	st	Z+,R16
	dbnzwh	R24,FIFOclrL
	ret

	;FIFO synchronous clear
FIFOclrSync:	brid	FIFOclrSyncS	;skip if global irq disabled
	lds	R16,timsk0
	andi	R16,1<<ocie0a
	breq	FIFOclrSyncS	;skip if FIFO output irq disabled
	;main sync fill
	ldi	R20,high(FIFOsize)
	inc	R20
	ldiw	R16,0	;0 sample
	ldi	R18,high(FIFO+FIFOsize)
	ldsw	R28,FIFOwr	;FIFO write pointer
FIFOclrSyncVS:	cli		;critical section begin
	ldsw	R24,FIFOrd	;FIFO read pointer
	sei		;critical section end
	subw	R24,R28	;free space
	brcc	PC+1+2
	subiw	R24,-FIFOsize	;if negative	
	cpi	R25,1	;at least 256Bytes there? 
	brlo	FIFOclrSyncVS	;no - try again
	;fill 256 bytes
	ldi	R19,256/4
FIFOclrSyncFDW:	stw	Y+,R16
	stw	Y+,R16	;store stereo sample
	cpi	R28,low(FIFO+FIFOsize)
	cpc	R29,R18
	brlo	PC+1+2
	ldiw	R28,FIFO
	dbnz	R19,FIFOclrSyncFDW	;next sample
	subi	R20,1
	brcc	FIFOclrSyncVS	;next 256 bytes
	;done
FIFOclrSyncS:	ret



;************************************************************************************************
;	Timeout timer
;************************************************************************************************
	;timeout interruption

Timer1ctc:	pusht	R16
	in	R17,sreg
	;
	lds	R16,TimOutDiv
	inc	R16
	cpi	R16,10
	brlo	PC+1+1+1+2
	clr	R16
	ldi	R18,0xFF
	sts	TimOut,R18
	sts	TimOutDiv,R16
	;
	out	sreg,R17
	popt	R16
	reti
	
	;timeout start
	.macro	ToRun
	lds	R16,timsk1
	sbr	R16,1<<ocie1a
	sts	timsk1,R16
	.endm
	
	;timeout stop
	.macro	ToStop
	lds	R16,timsk1
	cbr	R16,1<<ocie1a
	sts	timsk1,R16
	.endm
	
	;timeout set time (in [10ms], up to 33.500s @ XTAL=20MHz)
	.macro	ToTime
	ldi	R16,high(((XTAL/1024)*@0)/1000)
	sts	ocr1ah,R16
	ldi	R16,low(((XTAL/1024)*@0)/1000)
	sts	ocr1al,R16
	.endm
	
	;timeout reset
	.macro	ToRes
	clr	R16
	sts	tcnt1h,R16
	sts	tcnt1l,R16
	sts	TimOut,R16
	sts	TimOutDiv,R16
	.endm
	

;************************************************************************************************
;	Buttons timer
;************************************************************************************************
	;start button irq
	.macro	BtRun
	lds	R16,timsk2
	sbr	R16,1<<ocie2a
	sts	timsk2,R16
	.endm
	
	;stop button irq
	.macro	BtStop
	lds	R16,timsk2
	cbr	R16,1<<ocie2a
	sts	timsk2,R16
	.endm
	
	;set inhibition
	.macro	BtInh
	push	R16
	ldi	R16,@0
	sts	BtnInh,R16
	pop	R16
	.endm
	

	;timer 2 ocra irq
Timer2ctc:	push	R16
	in	R16,sreg
	push	R16
	
	BtStop
	sei		;enable golbal irq
	
	;do button stuff?
	lds	R16,BtnInh
	tst	R16
	rjne	Timer2ctcS	;skip if inhibited
	decs	R16,BtnDiv
	rjne	Timer2ctcS	;skip if not 0
	stsi	BtnDiv,4	;restore divisor const
	;do button stuff
	pusht	R17
	
	in	R17,ddra
	in	R18,porta	;store button port state
	outi	ddra,0x00	;inputs
	outi	porta,0xFF	;pull-ups
	sbi	dBCO,BCO	;set BCO as output
	adiw	R30,0	;nop,nop
	adiw	R30,0	;nop,nop
	lds	R16,BtnLast	;get last btn states
	com	R16
	in	R19,pina	;get actual states
	adiw	R30,0	;nop,nop
	cbi	dBCO,BCO	;BCO as input
	out	porta,R18
	out	ddra,R17	;restore button port state
	
	sts	BtnLast,R19	;store actual state as last one
	com	R19
	eor	R16,R19	;make differences
	and	R16,R19	;only key down events
	
	;some edge?
	push	R19
	tst	R16
	breq	Timer2ctcNE
	;yeah - reset buttons auto repeater
	ldi	R17,50
	sts	BtnAutoD,R17
	ldi	R17,0
	sts	BtnAutoS,R17
	
	rcall	BtnAct	;and do button action
		
	;no - check for button state
Timer2ctcNE:	pop	R19
	tst	R19
	breq	Timer2ctcNS
	;some button active
	decs	R16,BtnAutoD
	brne	Timer2ctcNS
	;delay done - reset delay counter with repeat rate
	lds	R16,BtnAutoR
	sts	BtnAutoD,R16
	
	lds	R16,BtnAutoS
	cpi	R16,0
	breq	PC+1+3
	stsi	BtnAutoS,1
	
	;and do button action
	mov	R16,R19
	rcall	BtnAct	;and do button action
	
Timer2ctcNS:	popt	R17
Timer2ctcS:	BtRun
	pop	R16
	out	sreg,R16
	pop	R16
	reti
	

	;------ button actions list ------
TBtnActList:	.dw	BtnAct0
	.dw	BtnAct1
	.dw	BtnAct2
	.dw	BtnAct3
	.dw	BtnAct4
	.dw	BtnAct5
	.dw	BtnAct6
	.dw	0


	;------ do button actions ------
BtnAct:	;select action (first non-zero bit from LSB)
	pushw	R30
	ldiw	R30,TBtnActList*2	;action list
	ldi	R17,8
BtnActAL:	lsr	R16
	brcc	BtnActNA
	;button pressed
	lpmw	R16,Z+
	movw	R30,R16
	adiw	R30,0
	breq	BtnActAD
	icall		;do action
	rjmp	BtnActAD
BtnActNA:	adiw	R30,2
	dbnz	R17,BtnActAL
	ldi	R16,30
BtnActAD:	popw	R30
	sts	BtnAutoR,R16	;store buttons auto repeater rate const.
	;
	ret


	;------ button actions ------
	;button 0 - file/folder selection (previous)
BtnAct0:	lds	R16,MODflag
	sbr	R16,1<<0|1<<4	;playback done
	cbr	R16,1<<1	;previous
	sts	MODflag,R16
	;
	ldi	R16,30
	ret

	;button 1 - file/folder selection (next) 
BtnAct1:	lds	R16,MODflag
	sbr	R16,1<<0|1<<1|1<<4	;playback done, next
	sts	MODflag,R16
	;
	ldi	R16,30
	ret

	;button 2 - pattern selection (previous)
BtnAct2:	lds	R16,MODpln
	cpi	R16,64
	breq	BtnAct2S
	lds	R16,MODpat
	subi	R16,2
	cpi	R16,-1
	brge	PC+1+1
	ldi	R16,-1
	sts	MODpat,R16
	stsi	MODpln,64	;previous pattern
BtnAct2S:	;
	ldi	R16,9
	ret
		
	;button 3 - pattern selection (next)
BtnAct3:	lds	R16,MODpat
	lds	R17,MODpatn
	dec	R17
	cp	R16,R17
	brlo	PC+1+1+2
	mov	R16,R17
	dec	R16
	sts	MODpat,R16
	stsi	MODpln,64	;new pattern
	;
	ldi	R16,9
	ret
	
	;button 4 - stop
BtnAct4:	lds	R16,MODflag
	bst	R16,2
	brtc	BtnAct4FLD
	;playing
	cbr	R16,1<<2	;stop
	rjmp	BtnAct4STP
	;stopped
BtnAct4FLD:	bst	R16,5
	brts	PC+1+1+1+1	;skip if no storage	
	ldi	R17,1<<3
	eor	R16,R17	;toggle folder selection flag	
BtnAct4STP:	sbr	R16,1<<0	;playback done
	sts	MODflag,R16
	;
	ldi	R16,30
	ret
	
	;button 5 - play
BtnAct5:	lds	R16,MODflag
	bst	R16,3
	brts	BtnAct5PLA
	;folder menu
	sbr	R16,1<<0|1<<3|1<<2	;playback done, play mode, play
	rjmp	BtnAct5FLD
	;play mode
BtnAct5PLA:	bst	R16,2
	brts	BtnAct5FLD
	sbr	R16,1<<0|1<<2	;playback done, play
BtnAct5FLD:	sts	MODflag,R16
	;
	ldi	R16,30
	ret
	
	;button 6 - load from COM
BtnAct6:	lds	R16,MODflag
	bst	R16,3
	brtc	BtnAct6N
	sbr	R16,1<<0|1<<6
	cbr	R16,1<<4
BtnAct6N:	sts	MODflag,R16
	;
	ldi	R16,30
	ret	
	
	

	
	
	
	
	
	
	

	;OTHER MODULES
	.include "usart.asm"
	.include "LCD.asm"
	.include "storage.asm"		
	
;************************************************************************************************
;	Initialization
;************************************************************************************************
Init:	;init ports/pins
	outi	ddrb,0b10111111
	cbi	pLCK,LCK	;LCD E=0
	cbi	pDWS,DWS	;DAC WS=0
	sbi	pRWR,RWR	;SRAM RW=1
	outi	ddrd,0b11111000
	sbi	pROE,ROE	;SRAM OE=1
	sbi	pRMA,RMA	;SRAM MA=1
	sbi	pRHA,RHA	;SRAM HA=1
	sbi	pIORD,IORD	;IDE IORD=1
	sbi	pIOWR,IOWR	;IDE IOWR=1
	cbi	pBCO,BCO	;BCO=0
	outi	ddra,0x00
	outi	ddrc,0x00
	
	
	;set stack
	outiw	sp,ramend
		
	;usart - rx, tx, 8bit, 1stop, no parity
	stsi	ucsr0a,0
	stsi	ucsr0b,1<<rxen0|1<<txen0
	stsi	ucsr0c,1<<ucsz00|1<<ucsz01
	stsiw	ubrr0,USRDIV
	
	;SPI - DAC output
	outi	spcr,1<<spe|1<<mstr
	outi	spsr,1<<spi2x	;10Mbit/s @ 20MHz
	stsiw	FIFOrd,FIFO
	stsiw	FIFOwr,FIFO	;reset FIFO pointers
	
	;timer 0 - FIFO output
	outi	tccr0a,1<<wgm01
	.if	SRDIV==8
	outi	tccr0b,1<<cs01	;/8
	.elif	SRDIV==64
	outi	tccr0b,1<<cs01|1<<cs00	;/64
	.else
	.error	"SAMPLING RATE OUT OF RANGE!!!"
	.endif
	outi	ocr0a,SROCR
	
	;timer 1 - timeout
	stsi	tccr1a,0
	stsi	tccr1b,1<<wgm12|1<<cs10|1<<cs12	;ocra, /1024
	
	;timer 2 - buttons
	stsi	tccr2a,1<<wgm21
	stsi	tccr2b,1<<cs20|1<<cs21|1<<cs22
	stsi	ocr2a,10*(XTAL/1024)/1000
	stsi	BtnDiv,4	;divisor const
	stsi	BtnInh,0	;not inhibited

	clrst	FileLen,R16
	
	;------ initialize LCD ------
	ldi	R16,20
	call	Delay10ms
	rcall	LCDinit
		
	;------ ext SRAM test ------
	call	SRAMtest
			
	;------ global IRQ enabled ------
	sei	
	
	;------ controll flags ------
	stsi	MODflag,0	;clear MOD flags
		
	;------ init storage ------
	;IDE
	.if	STORA!=0
	rcall	IDEinit	;initialize IDE device
	brts	STRGideOk
	.endif
	;no storage device
	sbsi	MODflag,1<<0|1<<2|1<<3|1<<5|1<<6	;no storage, play mode, play, playb. done, load COM  
	rjmp	MODplNew	;goto play menu
STRGideOk:	
	;FAT
	rcall	FATinit	;init FAT
	brts	STRGfatOk
	;invalid file system
	sbsi	MODflag,1<<0|1<<2|1<<3|1<<5|1<<6	;no storage, play mode, play, playb. done, load COM
	rjmp	MODplNew	;goto play menu
STRGfatOk:
		

	;--- enable buttons ---
	BtRun




	;-----------------------------------
	;------ folder selection menu ------
	;-----------------------------------
	stsi	FATrfld,0	;0. folder
	
	;scan root for folders 
FMstp:	rcall	FATscanRFld	;scan root
	ldiw	R30,FATsrt
	lds	R25,FATrfldc
	cpi	R25,0
	brne	FMstpNoFld
	;no folder there
	sbsi	MODflag,1<<0|1<<2|1<<3|1<<5|1<<6	;no storage, play mode, play, playb. done, load COM
	rjmp	MODplNew	;goto play menu
FMstpNoFld:	rcall	FATsortNLst	;sort root folders
	
	;scan selected folder for files
FMcrf:	ldiw	R30,FATsrt
	lds	R16,FATrfld
	rcall	FATscanFld	;scan selected folder
	
	;print files count in selected folder
	rcall	LCDclrL2
	lds	R16,FATflec
	rcall	LCDwrBCD3sz
	ldiw	R30,TSngs*2
	rcall	LCDwrStrFlash
	

	;--- folder menu controll loop ---
FMctrll:	BtRun
	ldi	R16,5
	call	Delay1ms
	BtStop
	
	lds	R16,MODflag	;ctrl flags
	bst	R16,3	;play mode?
	brtc	FMnp
	;play mode selected
	lds	R17,FATflec
	cpi	R17,1	;something to play?
	brsh	PC+1+1+2+1
	;no - stay in menu
	cbr	R16,1<<3|1<<0	;not play
	sts	MODflag,R16
	rjmp	PC+1+1
	rjmp	FMplay	;yeah - play selected folder
FMnp:	
	
	;--- change folder selection ---
	bst	R16,0	;playback done flag?
	brtc	FMncf
	;yeah - change folder
	cbr	R16,1<<0
	sts	MODflag,R16	;not playback done
	
	lds	R17,FATrfldc
	cpi	R17,2
	brlo	FMncf	;do nothing if 1 folder only 
	
	;direction?
	bst	R16,1	;up/down?
	brts	FMcfu
	;previous folder
	lds	R17,FATrfld
	cpi	R17,0
	brne	PC+1+2
	lds	R17,FATrfldc
	dec	R17
	sts	FATrfld,R17
	rjmp	FMcrf	;changed folder
	;next folder
FMcfu:	lds	R17,FATrfld
	inc	R17
	lds	R18,FATrfldc
	cp	R17,R18
	brlo	PC+1+1
	ldi	R17,0
	sts	FATrfld,R17
	rjmp	FMcrf	;changed folder
FMncf:	rjmp	FMctrll	;stay in menu
	

	
	;--- play mode selected ---	
FMplay:	ldiw	R30,FATsrt
	clr	R0
	lds	R16,FATrfld
	add	R30,R16
	adc	R31,R0
	ld	R16,Z
	sts	FATrflds,R16	;selected folder index (sorted)

	ldiw	R30,FATsrt
	lds	R25,FATflec
	rcall	FATsortNLst	;sort files in selected folder
	
	lds	R16,MODflag
	cbr	R16,1<<4	;not next song
	sbr	R16,1<<0	;playback done
	sts	MODflag,R16
	stsi	FATfle,0	;0. file selected
	
	


	;---------------------------------
	;------ file selection menu ------
	;---------------------------------
MODplNew:	BtRun
	ldi	R16,5
	call	Delay1ms
	BtStop


	lds	R16,MODflag
	bst	R16,3	;leave this menu?
	brts	PC+1+1+2+1
	;yeah
	cbr	R16,1<<0
	sts	MODflag,R16
	rjmp	FMstp	;goto folder selection
	

	bst	R16,0	;some other action to do?
	brtc	MODplNew
	;yeah
	cbr	R16,1<<0	;clear playback done flag
	sts	MODflag,R16
	
	bst	R16,4	;change song?
	brtc	MODplNewNPE	;no - skip file change
	;yeah - change it
	cbr	R16,1<<4	;clear it
	sts	MODflag,R16
	
	bst	R16,5
	brts	MODplNewNPE	;skip if no storage
	bst	R16,6
	brts	MODplNewNPE	;or COM load selected
	
	
	
	bst	R16,1	;direction?
	brts	MODplNewFP
	;previous file
	lds	R17,FATfle
	cpi	R17,0
	brne	PC+1+2
	lds	R17,FATflec
	dec	R17
	sts	FATfle,R17	;previous file
	rjmp	MODplNewNPE
	;next file
MODplNewFP:	lds	R18,FATfle
	inc	R18
	lds	R17,FATflec
	cp	R18,R17
	brlo	PC+1+1
	ldi	R18,0
	sts	FATfle,R18	;next file	
MODplNewNPE:	
	
	
	;--- initiate file playback ---
	stsiw	FIFOrd,FIFO
	stsiw	FIFOwr,FIFO	;reset FIFO pointers	
		
	;--- MOD song into SRAM ---
	lds	R16,MODflag
	bst	R16,6
	brts	MODplNewCOMF
	bst	R16,5	;no storage?
	brtc	MODplNewFLE
	;yeah - from COM port
	rjmp	MODplNewCOMNL
MODplNewCOMF:	cbr	R16,1<<6
	sts	MODflag,R16
	
MODplNewCOMLA:	rcall	LoadData	;time critical section - stop all irqs longer than few cycles!
	lds	R16,MODflag
	brts	MODplNewCOMLO
	bst	R16,5
	brtc	MODplNewFLE	;skip if COM load failed but storage ok
	;COM load failed, no storage
	ldst	R17,FileLen
	clr	R20
	tstt	R17,R20
	breq	MODplNewCOMLA
MODplNewCOMLO:	

	bst	R16,5
	brts	PC+1+1+2
	;storage ok
	sbr	R16,1<<2	;play if storage dev. ok 
	sts	MODflag,R16
	
	
	
MODplNewCOMNL:	rcall	MODprnNameE
	rjmp	MODplNewCOM
	
	;from file
MODplNewFLE:	clr	R18	;play - full song load
	lds	R16,MODflag
	bst	R16,2	;play/stop flag?
	brts	PC+1+1
	ser	R18	;stop - only song name printout
	lds	R16,FATrflds
	lds	R17,FATfle
	ldiw	R30,FATsrt
	rcall	FATrdFle	;load 'FATfle' file from 'FATrflds' folder	 
MODplNewCOM:
			
	;play?
	lds	R16,MODflag
	bst	R16,2
	rjtc	MODplNew	;no - file selection
	;yeah
	cbr	R16,1<<0|1<<4	;not playback done
	sts	MODflag,R16
		
	;clear status line
	rcall	LCDclrL2
	
	rcall	LoadMOD	;load MOD
	brts	MODplNewLDO
	;load failed
	lds	R16,MODflag
	sbr	R16,1<<0|1<<4	;forward, next song
	bst	R16,5
	bld	R16,6	;load again if no storage
	sts	MODflag,R16
	rjmp	MODplNew
	
MODplNewLDO:	rcall	InitMOD	;init MOD vars
		
	call	FIFOclr	;clear FIFO garbage
	DArun		;run DAC output
	
	BtRun		;allow buttons
	
	;config ports/signals
	outi	ddra,0xFF	;address port as output
	outi	ddrc,0x00	;data port as input
	outi	portc,0xFF	;with pull-ups
	cbi	pROE,ROE	;SRAM OE=0
	
	;main MOD player init
	stsi	FIFOtwr,0	;0 free bytes in FIFO
	stsi	MODncnt,1
	ldiw	R24,1
	pushw	R24	;init timing counters (on first note)
	

	
	
	;------ MOD PLAYBACK LOOP BEGIN ------
MODplLoop:	popw	R24	;load tick rate counter
	sbiw	R24,1
	rjne	MODplSTick
	ldsw	R24,MODtdiv
	pushw	R24	;store tick rate counter 
	;------ TICK CYCLE BEGIN ------	
	
	lds	R16,MODncnt
	bst	R16,0
	brts	PC+1+1
	rcall	LCDwrMODvu	;draw VUs
	
		
	;--- effects ---
	ldiw	R28,MODpbck
	clr	R2	;0
	lds	R24,MODchnn
MODplEL:	
	
	;RETRIGGER
	lds	R16,MODncnt
	cpi	R16,1
	brne	PC+1+1
	std	Y+meRT+0,R2	;new line in this tick, stop retrigger
	
	ldd	R16,Y+meRT+0
	cpi	R16,0
	breq	MODplENR	;skip if no effect active
		
	;get sample pointer
	ldd	R20,Y+15	;load actual sample id
	ldi	R16,MODsmplSize
	mul	R16,R20
	movw	R26,R0
	subiw	R26,-(MODsmpl-MODsmplSize)	;sample pointer
	
	ldd	R16,Y+meRT+0
	cpi	R16,0x09
	brne	MODplENRR
	;--- note retrigger
	cpi	R20,0	;sample id not specified yet?
	breq	MODplENR	;skip if no sample specified
	;some sample there
	
	ldd	R18,Y+meRT+1
	dec	R18
	brne	MODplENRRn
	;retrigger
	lds	R16,MODncnt
		
	;set note parameters
	adiw	R26,1
	ldtb	R18,X+	;get sample begin
	lddw	R16,Y+2	;resample_coeff
	ldi	R21,0
	sub	R21,R16
	sbc	R18,R17
	sbc	R19,R2
	sbc	R20,R2	;sample_begin-=resample_coeff	
	std	Y+4+0,R21	;store its fraction part		
	stdt	Y+4+1,R18	;store full part
	;retrigger note
	ldi	R16,0xFF
	std	Y+0,R16
	
	ldd	R18,Y+meRT+2
MODplENRRn:	std	Y+meRT+1,R18
	rjmp	MODplENR
	
MODplENRR:	cpi	R16,0x0C
	brne	MODplENRC
	;--- note cut off
	ldd	R18,Y+meRT+1
	dec	R18
	std	Y+meRT+1,R18
	brne	MODplENR
	;cut off note
	std	Y+meRT+0,R2	;no retrigger effect
	std	Y+1,R2	;0 volume
	rjmp	MODplENR
	
MODplENRC:	cpi	R16,0x0D
	brne	MODplENR
	;--- note delay
	ldd	R18,Y+meRT+1
	dec	R18
	std	Y+meRT+1,R18
	brne	MODplENR
	;trigger note
	std	Y+meRT+0,R2	;no retrigger effect

	ldd	R20,Y+meRT+2	;sample id
	lddw	R21,Y+meRT+3	;sample period
	ldi	R23,0	;no effect
	rcall	MODrPatPCNT	;trigger sample
MODplENR:	

	
	;VIBRATO
	ldd	R16,Y+meVI+0
	bst	R16,0
	rjtc	MODplEVBS
	;vibrato active
	lddw	R18,Y+meVI+5	;load waveform position [fx 8.8]
	lddw	R20,Y+meVI+2	;load waveform step [fx 8.8]
	mov	R30,R19
	clr	R31
	subiw	R30,-TMODsine32*2	;waveform table
	lpmw	R16,Z+	;read u1=u[t], u2=u[t+1] 
	addw	R18,R20	;add resample_step
	cpi	R19,32
	brlo	PC+1+1
	subi	R19,32	;check overrun
	stdw	Y+meVI+5,R18	;store position pointer
	
	;linear interpolation (much better results for higher frequencies)
	sub	R17,R16	;du=u2-u1
	brge	PC+1+1
	sub	R16,R18	;u1-=fraction if du<0, where fraction=R18 [fx 0.8]
	mul	R17,R18
	add	R16,R1	;u1+=(du*fraction)>>8
	
	;apply amplitude to waveform
	ldd	R17,Y+meVI+4	;amplitude
	clr	R18
	mulsu	R16,R17
	brcc	PC+1+1
	dec	R18	;u1*amplitude [fx 16.8]
	lddw	R20,Y+16	;load central period
	mov	R19,R0
	add	R20,R1
	adc	R21,R18	;sum it with freq. modulation delta [fx 16.8]
			
	;period limits
	brmi	PC+1+3+1
	cpiwh	R20,R16,MINT
	brpl	PC+1+3
	ldit	R19,MINT<<8	;check low limit
	cpiwh	R20,R16,MAXT
	brmi	PC+1+3
	ldit	R19,MAXT<<8	;check high limit
	;resampling coeff. (calculated in 24-bit precision instead of usual 16-bit)	
	ldit	R16,((3546895*256*2)/SREAL)*(128/4)
	lsrt	R19
	lsrt	R19	;period/=4
	call	div24u
	stdw	Y+2,R16	;store resampling coeff. [fx 8.8]
MODplEVBS:		
	
	;ARPEGGIO
	ldd	R16,Y+meAR+0
	bst	R16,0
	brtc	MODplARPS	;skip if not active
	;arpeggio active
	ldd	R20,Y+meAR+1	;load counter
	cpi	R20,0
	brne	MODplARPT0S
	;0. tone (base one)
	lddw	R18,Y+16
	rjmp	MODplARPTLD
MODplARPT0S:	cpi	R20,1
	brne	MODplARPT1S
	;1. tone
	lddw	R18,Y+meAR+2
	rjmp	MODplARPTLD
	;2. tone
MODplARPT1S:	lddw	R18,Y+meAR+4
MODplARPTLD:	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store resampling coeff
	;new tone
	inc	R20
	cpi	R20,3
	brlo	PC+1+1
	ldi	R20,0	;reset counter
	std	Y+meAR+1,R20	;store counter
MODplARPS:	

	lds	R16,MODncnt
	cpi	R16,1
	rjeq	MODplEPORTS	;skip if new line
	
	;VOLUME SLIDE
	ldd	R16,Y+meVS+0
	cpi	R16,0
	breq	MODplEVSS	;not active - skip
		
	;effect active
	ldd	R16,Y+meVS+1	;slide step
	ldd	R17,Y+1	;channel volume
	add	R17,R16	;add step
	cpi	R17,64
	brlt	MODplEVSLT
	;overflow
	ldi	R17,64
	std	Y+meVS+0,R2	;stop effect
	rjmp	MODplEVSGE	
MODplEVSLT:	cpi	R17,0
	brge	MODplEVSGE
	;underflow
	ldi	R17,0
	std	Y+meVS+0,R2	;stop effect
MODplEVSGE:	std	Y+1,R17	;store channel volume	
MODplEVSS:
	
	
	;PORTAMENTO
	ldd	R25,Y+meTP+0
	bst	R25,0	;activity bit
	rjtc	MODplEPORTS	;skip if not active
	;portamento active
	lddw	R18,Y+16	;load period
	ldd	R16,Y+meTP+1	;portamento step
	ldi	R17,0
	bst	R25,2	;tone-portamento bit
	brtc	MODplEPORTN
	
	;tone-portamento active (3)
	lddw	R20,Y+18	;load final period
	bst	R25,1	;portamento direction
	brtc	MODplEPORTTU
	;period down
	subw	R18,R16	;sub step
	brpl	PC+1+2
	ldiw	R18,MINT	;check underflow
	cpiwh	R18,R22,MINT
	brsh	PC+1+2
	ldiw	R18,MINT	;check lowest period	
	cpw	R18,R20
	brsh	MODplEPORTHOK	
	movw	R18,R20	;final reached
	ldd	R16,Z+0
	cbr	R16,1<<0
	std	Y+meTP+0,R16	;stop effect
	rjmp	MODplEPORTHOK
	;period up
MODplEPORTTU:	addw	R18,R16	;add step
	cpw	R18,R20
	brlo	MODplEPORTHOK	
	movw	R18,R20	;final reached
	ldd	R16,Y+meTP+0
	cbr	R16,1<<0
	std	Y+meTP+0,R16	;stop effect
	rjmp	MODplEPORTHOK
 	
	
	;portamento active (1/2)
MODplEPORTN:	bst	R25,1	;portamento direction
	brtc	MODplEPORTU
	;portamento up
	negw	R16	;negative step
MODplEPORTU:	addw	R18,R16	;add step to period
	cpiwh	R18,R16,MINT
	brsh	MODplEPORTLOK
	;to low period
	ldiw	R18,MINT
	std	Y+meTP+0,R2	;stop effect
	rjmp	MODplEPORTHOK	
MODplEPORTLOK:	cpiwh	R18,R16,MAXT
	brlo	MODplEPORTHOK
	;to high period
	ldiw	R18,MAXT
	std	Y+meTP+0,R2	;stop effect
MODplEPORTHOK:	stdw	Y+16,R18	;store period
	;calculate new resampling coeff
	;ldiw	R16,256*3546895/SREAL
	;rcall	div16u
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store resampling coeff	
MODplEPORTS:
	
	adiw	R28,MODpbckSize	;skip playback channel
	djnz	R24,MODplEL
	
	
	

	;------ TICK CYCLE END ------
	decs	R24,MODncnt
	rjne	MODplSNote
	lds	R24,MODspd
	sts	MODncnt,R24	;store note rate counter
	;------ NOTE CYCLE BEGIN ------
	
	
	;stop some effects
	ldiw	R28,MODpbck
	clr	R0
	lds	R24,MODchnn
MODplNCCE:	;stop volume slide effect
	std	Y+meVS+0,R0
	;stop portamento effect
	ldd	R16,Y+meTP+0
	cbr	R16,1<<0
	std	Y+meTP+0,R16
	;stop retrigger effects
	std	Y+meRT+0,R0	;activity
	;
	adiw	R28,MODpbckSize	;next channel
	dbnz	R24,MODplNCCE

	

	;playback controll
	lds	R16,MODflag
	bst	R16,0
	brts	MODplPD	;forced end
	
	;read line
	rcall	MODrPat	;read pattern line
	brtc	MODplNEX
	;playback done - song end
	lds	R16,MODflag
	sbr	R16,1<<0|1<<1|1<<4	;forward, next song
	bst	R16,5
	bld	R16,6	;load COM again if no storage
	sts	MODflag,R16
MODplPD:	;playback done - forced end
	popw	R24	;pop tick counter
	;config ports/signals
	outi	ddra,0x00	;inputs
	sbi	pROE,ROE	;SRAM OE=1
	call	FIFOclrSync	;shut up DAC FIFO
	DAstop		;stop DAC	
	rjmp	MODplNew
MODplNEX:	

	;write status (line/pattern/total patterns)
	rcall	LCDwrMODln
		
	;test for speed/tempo modification
	lds	R16,MODtimc
	bst	R16,0	;speed changed?
	brtc	MODplSMN
	;yeah - reset note cycle counter
	lds	R16,MODspd
	sts	MODncnt,R16
	cbs	MODtimc,0	;clear flag
	rjmp	MODplSTN
MODplSMN:	bst	R16,1	;tempo changed?
	brtc	MODplSTN
	;yeah - reset tick cycle counter
	popw	R16	;load tick cycle counter
	ldsw	R16,MODtdiv
	pushw	R16	;store tick cycle counter
	cbs	MODtimc,1
MODplSTN:


	;------ NOTE CYCLE END ------
MODplSNote:	rjmp	PC+1+2
MODplSTick:	pushw	R24	;store note rate counter
	;------ SAMPLE CYCLE BEGIN ------
	
	ldsw	R28,FIFOwr	;FIFO write pointer
	lds	R24,FIFOtwr	;load free samples count
	cpi	R24,0
	brne	MODplTWR	;skip if some place in FIFO
	;no place in FIFO
	
	;wait for free space in DAC FIFO
	cbi	portb,4	
MODplFF:	cli		;critical section begin
	ldsw	R24,FIFOrd	;FIFO read pointer
	sei		;critical section end
	subw	R24,R28
	brcc	MODplFP
	subiw	R24,-FIFOsize	;negative	
MODplFP:	breq	MODplFF	;try again if no space at all
	;some space there
	lsrw	R24
	lsrw	R24	;in samples
	tst	R25
	breq	PC+1+1
	ldi	R24,256	;max 256 samples at once
				
	sbi	portb,4		
MODplTWR:	dec	R24	;-sample
	sts	FIFOtwr,R24	;store it
	
	
	
	.if	DRAWVU==1
	lds	R26,MODchnn
	clr	R27
	subiw	R26,-MODbars
	.endif
	ldiw	R30,MODpbck	;playback structure
	clr	R8	;0
	ldi	R21,1	;1
	.if	STEREO==1
	clrw	R14
	.endif
	clrw	R12
	;ldi	R24,4	;4 channels
	lds	R24,MODchnn
	;for each channel
MODplMPC:	ldd	R16,Z+0	;playback activity
	cpi	R16,0	;playback stopped?
	.if	DRAWVU==1
	brne	PC+1+1+1
	sbiw	R26,1	;previous VU channel
	rjmp	MODplMPSC	;yeah - skip channel
	.else
	rjeq	MODplMPSC
	.endif
	lddw	R6,Z+2	;load resampling coeff.
	lddd	R2,Z+4	;load sample data position
	lddt	R9,Z+8	;load sample data end
	
	addw	R2,R6
	adc	R4,R8
	adc	R5,R8	;add resample coeff
	cpt	R3,R9	;test for sample end
	brlo	MODplNE
	;end, test for loop
	ldd	R16,Z+14	;loop activity
	cpi	R16,0
	breq	MODplMPE
	;loop active - reset pointer
	lddt	R16,Z+11	;load offset (relative!)
	subt	R3,R16
	rjmp	MODplNE		
	;no loop
MODplMPE:	std	Z+0,R8	;stop playback!
	stdd	Z+4,R2	;store sample data position
	.if	DRAWVU==1
	sbiw	R26,1	;previous VU channel
	.endif
	rjmp	MODplMPSC	;skip rest of channel!		
MODplNE:	stdd	Z+4,R2	;store sample data position
		
	;now we have valid pointer in R3:R4:R5
	;read sample
	out	porta,R5
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	out	porta,R4
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	out	porta,R3
	.if	HCOC==1
	nop
	.endif
	.if	LINRES==1
	add	R3,R21
	adc	R4,R8
	adc	R5,R8	;pointer to next sample
	.else
	ldd	R20,Z+1	;load volume
	bst	R24,1	;T=1/0 for left/right channel
	.endif
	in	R16,pinc	;u1=u[t], where t=R3:R4:R5
		
	;read next sample	
	.if	LINRES==1
	out	porta,R5
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	out	porta,R4
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	out	porta,R3
	.if	HCOC==1
	nop
	.endif
	ldd	R20,Z+1	;load volume
	bst	R24,1	;T=1/0 for left/right channel
	in	R17,pinc	;u2=u[t+1], where t=R3:R4:R5
	
	;linear interpolation ;-D
	sub	R17,R16	;du=u2-u1
	brge	PC+1+1
	sub	R16,R2	;u1-=fraction if du<0, where fraction=R2 [fx 0.8]
	mul	R17,R2
	add	R16,R1	;u1+=(du*fraction)>>8
	.endif
	
	;apply volume
	.if	STEREO==1
	.if	MAMPLD==1
	fmulsu	R16,R20	;apply volume: R0:R1=2*u1*volume (max +16256/-16384), where volume=R20 [0-64]
	.else
	mulsu	R16,R20	;apply volume: R0:R1=u1*volume (max +8128/-8192), where volume=R20 [0-64]
	.endif
	.else
	mulsu	R16,R20	;apply volume: R0:R1=u1*volume (max +8128/-8192), where volume=R20 [0-64]
	.if	MAMPLD==0
	asrw	R0	;apply volume: R0:R1=u1*volume (max +4064/-4096), where volume=R20 [0-64]	
	.endif
	.endif
	
	.if	DRAWVU==1
	;peak detection
	ld	R17,-X
	cp	R1,R17
	brlt	PC+1+1
	st	X,R1
	.endif
	
	;mixing - stereo channel selection (LRRLLRR ...)
	.if	STEREO==1
	brts	PC+1+2+1
	addw	R14,R0
	rjmp	PC+1+2
	addw	R12,R0
	.else
	addw	R12,R0
	.endif
	
MODplMPSC:	adiw	R30,MODpbckSize	;to next channel
	djnz	R24,MODplMPC
	
	.if	STEREO==1
	st	Y+,R15
	st	Y+,R14
	st	Y+,R13
	st	Y+,R12
	.else
	st	Y+,R13
	st	Y+,R12
	st	Y+,R13
	st	Y+,R12
	.endif
	
	cpiwh	R28,R16,FIFO+FIFOsize
	brlo	PC+1+2
	ldiw	R28,FIFO	
	stsw	FIFOwr,R28
	

	;------ SAMPLE CYCLE END------
	;------ MOD PLAYBACK LOOP END ------
	rjmp	MODplLoop	;playback loop	
	





;------------------------------------------------------------------------------------------------
;	MOD - read pattern line
;------------------------------------------------------------------------------------------------
MODrPat:	lds	R16,MODpln	;load line id
	cpi	R16,64
	brne	MODrPatNNL
	;new line
	lds	R16,MODpnln	;reset line counter
	sts	MODpln,R16	;store line counter
	stsi	MODpnln,0
	
	;new pattern
	incs	R16,MODpat	;next pattern
	lds	R17,MODpatn	;song len
	cp	R16,R17
	brlo	MODrPatNNP
	;all patterns played - stop flag
	set		;finish
	ret		;finish
MODrPatNNP:	;------ generate actual line pointer (SRAM) ------
	;first physical pattern pointer
	ldiw	R16,20+31*30+1+1+128+4	;title+chns*30+some_garbage+mahoney@kaktus
	lds	R18,MODinsn
	cpi	R18,31
	breq	MODrPat31i
	ldiw	R16,20+15*30+1+1+128	;title+chns*30+some_garbage
MODrPat31i:	;now add the rest
	ldiw	R30,MODpatl	;pattern list
	lds	R19,MODpat
	clr	R0
	add	R30,R19
	adc	R31,R0	;actual pattern pointer
	ld	R19,Z	;actual pattern id
	lds	R20,MODchnn
	mul	R19,R20	;pattern_id*pat_size/256
	clr	R18
	addw	R17,R0	;+=pattern_id*1024
	stst	MODpatp,R16	;store it
MODrPatNNL:	;------ main pattern reading ------
	ldst	R2,MODpatp	;load note pointer (SRAM)
	clr	R5
	inc 	R5
	clrw	R6	;SRAM step=+1
	
	ldiw	R28,MODpbck	;playback structure
	lds	R25,MODchnn	;channels
	;for each channel
MODrPatNC:	rcall	RamGetByteA
	mov	R20,R16
	andi	R20,0xF0
	mov	R22,R16
	andi	R22,0x0F	;period hi
	rcall	RamGetByteA
	mov	R21,R16	;period lo
	rcall	RamGetByteA
	mov	R23,R16
	andi	R23,0x0F	;effect id
	swap	R16
	andi	R16,0x0F
	or	R20,R16	;sample id
	rcall	RamGetByteA
	mov	R24,R16	;effect parameter
	
	
	;check for note delay effect
	cpi	R23,0x0E
	brne	MODrPatNND
	mov	R16,R24
	andi	R16,0xF0
	cpi	R16,0xD0
	brne	MODrPatNND
	;note delay effect
	mov	R16,R24
	andi	R16,0x0F
	breq	MODrPatNND	;skip if zero delay
	;non-zero delay
	std	Y+meRT+1,R16	;store delay
	ldi	R16,0x0D
	std	Y+meRT+0,R16	;store effect code
	std	Y+meRT+2,R20	;store sample id value
	stdw	Y+meRT+3,R21	;store sample period
	rjmp	MODrPatND	;skip note trigger
MODrPatNND:	
	
	;--- process basic note ---
	rcall	MODrPatPCNT
MODrPatND:			
	
	;------ effects ------
	
	;--- stop vibrato if no vibrato effect done (0x04/0x06) ---
	;--- and reset resampl. coeff. to central freq. ---
	cpi	R23,0x04
	breq	MODrPatNSV
	cpi	R23,0x06
	breq	MODrPatNSV
	;yeah - stop it
		
	ldd	R16,Y+meVI+0
	bst	R16,0
	brtc	MODrPatNSV	;not running, leave
	;running - stop it
	cbr	R16,1<<0
	std	Y+meVI+0,R16
	;and restore resampl. coeff.
	lddw	R18,Y+16	;load central period
	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store resampling coeff
MODrPatNSV:		
	
	
	;--- main effect code ---
	cpi	R23,0x00
	rjne	MODrPatNE0
	;0 - Arpeggio
	
	cpi	R24,0
	rjeq	MODrPatNE0	;leave if no parameter
	
	lddw	R21,Y+16	;load last used period
	cpiwh	R21,R16,1712+1
	rjsh	MODrPatNE0	;leave if period higher than C-0
		
	;identify current note
	ldiw	R30,TMODarpHS*2
	ldi	R20,0	;0 note defaultly
	
MODrPatEFL:	lpmw	R16,Z+	;read first note
	cpw	R16,R21	;compare period with table item
	brsh	MODrPatEFD	;note identified if table item same or higher than period   
	;continue
	inc	R20	;next note
	rjmp	MODrPatEFL
		
	
	;create effect pointer
MODrPatEFD:	ldd	R16,Y+meAR+0	;status byte
	bst	R16,0	;activity bit
	brts	MODrPatE0R	;skip if active
	;not running now
	std	Y+meAR+1,R7	;clear tone counter
	sbr	R16,1<<0	;active
	std	Y+meAR+0,R16	;store it
MODrPatE0R:	
	tst	R11
	breq	MODrPatENN
	;new note
	std	Y+meAR+1,R7	;clear tone counter
MODrPatENN:		
	;R20 is actual note id 
	;calculate 1. tone
	mov	R17,R24
	swap	R17
	andi	R17,0x0F
	mov	R30,R20
	sub	R30,R17	;1. tone id
	brge	PC+1+1
	ldi	R30,0	;arpeggio out of valid range
	clr	R31
	lsl	R30
	subiw	R30,-TMODarpHS*2
	lpmw	R16,Z+	;read 1. tone period
	stdw	Y+meAR+2,R16	;store 1. tone period
	
			
	;calculate 2. tone
	mov	R17,R24
	andi	R17,0x0F
	mov	R30,R20
	sub	R30,R17	;2. tone id
	brge	PC+1+1
	ldi	R30,0	;arpeggio out of valid range
	clr	R31
	lsl	R30
	subiw	R30,-TMODarpHS*2
	lpmw	R16,Z+	;read 2. tone period
	stdw	Y+meAR+4,R16	;store 2. tone period
	rjmp	MODrPatEDONE
	
	;arpeggio not active - stop it
	;create effect pointer
MODrPatNE0:	ldd	R16,Y+meAR+0
	bst	R16,0
	cbr	R16,1<<0	;stop
	std	Y+meAR+0,R16
	brtc	MODrPatNE0SFR
	;stop
	tst	R11
	brne	MODrPatNE0SFR	
	;no new note - restore last pariod
	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	lddw	R18,Y+16
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store resample coeff [fx 8.8] 	
MODrPatNE0SFR:	

	cpi	R23,0x01
	brne	MODrPatNE1
	;1 - Portamento up
	ldi	R16,1<<0|1<<1|0<<2	;portamento active, up
	std	Y+meTP+0,R16	;store it
	std	Y+meTP+1,R24	;store step
	rjmp	MODrPatEDONE

MODrPatNE1:	cpi	R23,0x02
	brne	MODrPatNE2
	;2 - Portamento down
	ldi	R16,1<<0|0<<1|0<<2	;portamento active, down
	std	Y+meTP+0,R16	;store it
	std	Y+meTP+1,R24	;store step
	rjmp	MODrPatEDONE

MODrPatNE2:	cpi	R23,0x03
	brne	MODrPatNE3
	;3 - Tone portamento
				
	cpi	R24,0
	brne	MODrPatE3N
	;no parameter
	ldd	R24,Y+meTP+1			
MODrPatE3N:	;some parameter specified
	std	Y+meTP+1,R24	;store slide step
	
	lddw	R21,Y+18	;final period
	lddw	R16,Y+16	;load channel period
	cpw	R16,R21	;compare with final period
	brne	MODrPatE3TSM
	;periods are same - nothing to do
	std	Y+meTP+0,R7	;stop effect	
	rjmp	MODrPatEDONE
MODrPatE3TSM:	brlo	MODrPatE3TLO
	;actual higher than final
	ldi	R16,1<<0|1<<1|1<<2
	std	Y+meTP+0,R16
	rjmp	MODrPatEDONE
	;actual lower than final
MODrPatE3TLO:	ldi	R16,1<<0|0<<1|1<<2
	std	Y+meTP+0,R16
	rjmp	MODrPatEDONE	
	

MODrPatNE3:	cpi	R23,0x04
	rjne	MODrPatNE4
	;4 - Vibrato
	rcall	MODrPatE4R
	rjmp	MODrPatEDONE
	
MODrPatE4R:	lddw	R21,Y+16	;load last used period
	
	mov	R16,R24
	andi	R16,0x0F
	brne	MODrPatE4NPA
	;missing amplitude parameter
	ldd	R16,Y+meVI+1
	andi	R16,0x0F
	breq	MODrPatE4D	;no loaded yet
MODrPatE4NPA:	or	R24,R16	;add to parameter

	mov	R16,R24
	andi	R16,0xF0
	brne	MODrPatE4NPF
	;missing amplitude parameter
	ldd	R16,Y+meVI+1
	andi	R16,0xF0
	breq	MODrPatE4D	;no loaded yet
MODrPatE4NPF:	or	R24,R16	;add to parameter
	std	Y+meVI+1,R24	;store last parameter
	
	;calculate vibrato waveform step 
	mov	R17,R24
	swap	R17
	clr	R16
	andi	R17,0x0F	;x=cycles*64 per tick
	lsrw	R16	;vib_step_per_tick=cycles*samples_per_period/64 [fx 8.8], where samples_per_period=32
	stdw	Y+meVI+2,R16	;store vib_step_per_tick [fx 8.8]
		
	;calculate vibrato amplitude
	mov	R17,R24
	andi	R17,0x0F
	lsl	R17
	lsl	R17	;param*4 [fx 0.8] (same as in MpTracker) 
	std	Y+meVI+4,R17	;store amplitude
		
	ldd	R16,Y+meVI+0
	bst	R16,0	;activity bit
	sbr	R16,1<<0
	std	Y+meVI+0,R16	;vibrato active
	
	brts	MODrPatE4SRTG
	;first run - retrigger
	ldiw	R16,0
	stdw	Y+meVI+5,R16	;store waveform position [fx 8.8]
MODrPatE4SRTG:	
MODrPatE4D:	ret
	
MODrPatNE4:	cpi	R23,5
	brne	MODrPatNE5
	;5 - Continue Tone-Portamento and volume slide (0x0A)
	
	;first volume slide
	rcall	MODrPatEAR
	
	;now do tone-portamento
MODrPatE5D:	ldi	R24,0	;no parameter
	ldi	R23,0x03	;tone-portamento
	rjmp	MODrPatNE2	;run it	
	


MODrPatNE5:	cpi	R23,0x06
	brne	MODrPatNE6
	;6 - continue vibrato and volume slide
	;first volume lisde
	rcall	MODrPatEAR

	;then vibrato
	ldi	R24,0	;no parameter
	rcall	MODrPatE4R
	rjmp	MODrPatEDONE
	
		

MODrPatNE6:	cpi	R23,0x09
	brne	MODrPatEN9
	;9 - Set sample offset
	cpi	R24,0
	brne	MODrPatE9YP	;skip if some argument
	;no argument, use last one
	ldd	R24,Y+mePS+0	;load it
	cpi	R24,0
	rjeq	MODrPatEDONE	;leave if 0 yet (indefined)
MODrPatE9YP:	std	Y+mePS+0,R24	;store actual arg. as last one
	
	ldd	R16,Y+0
	cpi	R16,0
	rjeq	MODrPatEDONE	;leave if no sample playing now
	ldd	R16,Y+15	;last instrument (actual)
	cpi	R16,0
	rjeq	MODrPatEDONE	;leave if no sample specified yet
	ldi	R17,MODsmplSize
	mul	R16,R17
	movw	R30,R0
	subiw	R30,-(MODsmpl-MODsmplSize)	;sample	
	
	lddt	R16,Z+1	;sample start
	add	R17,R24
	adc	R18,R7	;add offset*0x100
	lddt	R19,Z+4	;sample end
	cpt	R16,R19	;is offset in valid range?
	rjsh	MODrPatEDONE	;no - leave
	;ok, sample_begin+offset*0x100<sample_end
	lddw	R19,Y+2	;load resampling coeff 
	clr	R15	;0 fraction part
	subw	R15,R19
	sbc	R17,R7
	sbc	R18,R7	;-=resampling_coeff
	stdd	Y+4,R15	;store new sample position	
	rjmp	MODrPatEDONE
	
	

MODrPatEN9:	cpi	R23,0x0A
	brne	MODrPatENA
	;A - volume slide
	;effect pointer 
	rcall	MODrPatEAR
	rjmp	MODrPatEDONE
	

MODrPatEAR:	mov	R16,R24
	andi	R16,0x0F
	breq	MODrPatEAX5
	;y!=0 -> volume down
	neg	R16
	std	Y+meVS+1,R16	;store slide speed
	ldi	R16,0xFF
	std	Y+meVS+0,R16	;slide active	
	rjmp	MODrPatEAE
	;y==0 -> volume up
MODrPatEAX5:	mov	R16,R24
	swap	R16
	andi	R16,0x0F
	cpi	R16,0
	breq	MODrPatEAE	;skip if zero
	std	Y+meVS+1,R16	;store slide speed
	ldi	R16,0xFF
	std	Y+meVS+0,R16	;slide active
MODrPatEAE:	ret
	
	
MODrPatENA:	cpi	R23,0x0C
	brne	MODrPatENC
	;C - set volume
	cpi	R24,64
	brlo	MODrPatECVC
	ldi	R24,64	;cut volume upto 64 because some composers using 0-64 as decadic value 
			;(in fact it's coded as hexa, valid volumes are {C00,C01,...,C3F,C40})
MODrPatECVC:	std	Y+1,R24	;store volume
	rjmp	MODrPatEDONE

MODrPatENC:	cpi	R23,0x0D
	brne	MODrPatEND
	;D - pattern break
;	mov	R16,R24
;	swap	R16
;	andi	R16,0x0F
;	ldi	R17,10
;	mul	R16,R17
;	andi	R24,0x0F
;	add	R24,R0	;new_line=x*10+y (from some reason it's coded as BCD in this effect)
	stsi	MODpln,63	;last line
	cpi	R24,64
	rjsh	MODrPatEDONE
	sts	MODpnln,R24	;new line id		
	rjmp	MODrPatEDONE


MODrPatEND:	cpi	R23,0x0E
	rjne	MODrPatENE
	;E - Extended effects
	mov	R23,R24
	swap	R23
	andi	R23,0x0F	;subeffect code
	andi	R24,0x0F	;subeffect parameter
	
	cpi	R23,0x01
	brne	MODrPatEEN1
	;E1 - Fineslide up
	lddw	R18,Y+16	;load period
	sub	R18,R24
	sbc	R19,R7	;decrease period once
	cpiwh	R18,R16,MINT
	brsh	MODrPatEEN1O
	;to low - cut off
	ldiw	R18,MINT
MODrPatEEN1O:	stdw	Y+16,R18	;store period
	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store new resample coeff
	rjmp	MODrPatEDONE
		
MODrPatEEN1:	cpi	R23,0x02
	brne	MODrPatEEN2
	;E2 - Fineslide down
	lddw	R18,Y+16	;load period
	add	R18,R24
	adc	R19,R7	;increase period once
	cpiwh	R18,R16,MAXT
	brlo	MODrPatEEN2O
	;to high - cut off
	ldiw	R18,MAXT
MODrPatEEN2O:	stdw	Y+16,R18	;store period
	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store new resample coeff
	rjmp	MODrPatEDONE

MODrPatEEN2:	cpi	R23,0x09
	brne	MODrPatEEN9
	;E9 - Sample retrigger
			
	cpi	R24,0
	brne	PC+1+1
	ldi	R24,1	;1 if x==0
	std	Y+meRT+1,R24	;store retrigger counter
	std	Y+meRT+2,R24	;repeat value
	std	Y+meRT+0,R23	;store effect code
	rjmp	MODrPatEDONE
	
MODrPatEEN9:	cpi	R23,0x0A
	brne	MODrPatEENA
	;EA - Fine volume slide up
	ldd	R16,Y+1	;load volume
	add	R16,R24	;increment volume once
	cpi	R16,64
	brlo	MODrPatEEAO
	;too high - cut off
	ldi	R16,64
MODrPatEEAO:	std	Y+1,R16	;store volume back
	rjmp	MODrPatEDONE		

MODrPatEENA:	cpi	R23,0x0B
	brne	MODrPatEENB	
	;EB - Fine volume slide down
	ldd	R16,Y+1	;load volume
	sub	R16,R24	;increment volume once
	brge	MODrPatEEBO
	;too low - cut off
	ldi	R16,0
MODrPatEEBO:	std	Y+1,R16	;store volume back	
	rjmp	MODrPatEDONE

MODrPatEENB:	cpi	R23,0x0C
	brne	MODrPatEENC
	;EC - Note cut off
			
	cpi	R24,0
	brne	MODrPatEEC
	;immeadiate cut off
	std	Y+1,R7	;0 volume
	rjmp	MODrPatEDONE
MODrPatEEC:	;delayed cut off
	std	Y+meRT+1,R24	;store retrigger counter
	std	Y+meRT+0,R23	;store effect code
	rjmp	MODrPatEDONE 
MODrPatEENC:	rjmp	MODrPatEDONE


MODrPatENE:	cpi	R23,0x0F
	brne	MODrPatENF
	;F - set speed/tempo
	cpi	R24,0
	breq	MODrPatEDONE	;invalid parameter - abort
	cpi	R24,32
	brsh	MODrPatEFT
	;set speed
	sts	MODspd,R24
	sbs	MODtimc,0	;speed modification flag
	rjmp	MODrPatEDONE
	;set tempo
MODrPatEFT:	sts	MODbpm,R24	;store beats per minute
	;recalculate tick divisor
	ldit	R16,5*SREAL	;5*sample_rate(real one!)
	mov	R19,R24
	clrw	R20
	lslw	R19	;2*bpm
	rcall	div24u	;R16:R17=sample_rate/(0.4*bpm)
	stsw	MODtdiv,R16	;store it
	sbs	MODtimc,1	;tempo modification flag
	rjmp	MODrPatEDONE
MODrPatENF:	

	;effects done
MODrPatEDONE:	
	
	
	adiw	R28,MODpbckSize	;next channel 
	djnz	R25,MODrPatNC
	stst	MODpatp,R2	;store note pointer (SRAM)

	incs	R16,MODpln	;line++
	clt		;not finish	
	ret
	
	
	;------ process note ------
	;inputs:
	; R20 - sample id
	; R21:R22 - sample period
	; R23 - effect id
	; Y - playback structure pointer for actual channel
	;outputs:
	; R11 - non-zero if note on line
MODrPatPCNT:	cpi	R20,0	;is there sample id?
	breq	MODrPatNN
	;note id non zero
	std	Y+15,R20	;store sample id
	;pointer to sample R20 
	ldi	R16,MODsmplSize
	mul	R16,R20
	movw	R30,R0
	subiw	R30,-(MODsmpl-MODsmplSize)	;sample	
	
	;reset volume
	ldd	R16,Z+0	;sample volume
	std	Y+1,R16	;store it
MODrPatNN:	
	
	clr	R11	;no note on this line
	cp	R21,R11
	cpc	R22,R11	;is there sample period?
	rjeq	MODrPatNP
	;yeah, some period specified
	ldd	R20,Y+15	;load actual sample id
	cpi	R20,0	;not specified yet?
	brne	MODrPatNP1	
	;yeah - leave
	stdw	Y+16,R21	;store period
	rjmp	MODrPatNP
MODrPatNP1:	;some sample specified
	;pointer to sample R20
	ldi	R16,MODsmplSize
	mul	R16,R20
	movw	R30,R0
	subiw	R30,-(MODsmpl-MODsmplSize)	;sample
	
	
	;test for tone-portamento
	set
	cpi	R23,0x03
	breq	MODrPatTP3
	cpi	R23,0x05
	breq	MODrPatTP3
	clt
MODrPatTP3:
	
	stdw	Y+18,R21	;store tone-port. final period
	
	;set note playback rate
	brtc	MODrPatTPN
	;tone-portamento active
	lddw	R18,Y+16	;load last period instead of new specified!!!
	rjmp	MODrPatTP3D		
MODrPatTPN:	stdw	Y+16,R21	;store new period
	mov	R18,R21
	mov	R19,R22	;new T	
MODrPatTP3D:	;ldiw	R16,256*3546895/SREAL	;PAL/(2*SREAL) [fx 8.8]
	;rcall	div16u	;PAL/(2*SREAL)/T
	T2RC		;PAL/(2*SREAL)/T
	stdw	Y+2,R16	;store resample coeff [fx 8.8] 	

	;set note parameters
	brtc	MODrPatPAS2
	ldd	R26,Y+0	;playback activity?
	cpi	R26,0x00
	brne	MODrPatPAS1
MODrPatPAS2:	lddt	R18,Z+1	;sample begin
	ldi	R26,0	;its fraction part (R26:R18:R19:R20 [fx 24.8])
	sub	R26,R16
	sbc	R18,R17
	sbc	R19,R11
	sbc	R20,R11	;sample_begin-=resample_coeff	
	std	Y+4+0,R26	;store its fraction part		
	stdt	Y+4+1,R18	;store full part
MODrPatPAS1:	lddt	R16,Z+4	;sample end
	stdt	Y+8,R16	;store it
	lddt	R16,Z+7	;sample repeat
	stdt	Y+11,R16	;store it
	ldd	R16,Z+10	;sample loop activity
	std	Y+14,R16	;store it
	;start playback
	ldi	R16,0xFF
	std	Y+0,R16	;playback activity
	inc	R11	;note on this line
MODrPatNP:	ret


;------------------------------------------------------------------------------------------------
;	print MOD name to line 2 from Y pointed file data (only first sector) 
;------------------------------------------------------------------------------------------------
MODprnName:	ldiw	R30,MODname
	ldi	R17,20
	ldi	R16,' '
	st	Z+,R16	;clear string first
	dbnz	R17,PC-1-1
	;
	ldiw	R30,MODname
	ldi	R17,20
	ld	R16,Y+
	cpi	R16,0
	breq	PC+1+1+1+1	;name end reached
	st	Z+,R16	;store char
	dbnz	R17,PC-1-1-1-1-1
	;
	call	LCDposL2
	ldiw	R30,MODname
	ldi	R17,20
	call	LCDwrStrRam
	ret
	
;------------------------------------------------------------------------------------------------
;	print MOD name to line 2 from external SRAM 
;------------------------------------------------------------------------------------------------
MODprnNameE:	;config ports/signals
	BtInh	1
	outi	ddra,0xFF	;outputs
	outi	portc,0xFF	;pull-up
	cbi	pROE,ROE	;SRAM OE=0

	ldiw	R30,MODname
	ldi	R17,20
	ldi	R16,' '
	st	Z+,R16	;clear string first
	dbnz	R17,PC-1-1
		
	;SRAM address steping (R5:R6:R7)
	clr	R5
	inc	R5
	clrw	R6	;+1
	;SRAM address (R2:R3:R4)=0
	clrt	R2
	
	ldiw	R30,MODname
	ldi	R17,20
	rcall	RamGetByteA
	cpi	R16,0
	breq	PC+1+1+1+1	;name end reached
	st	Z+,R16	;store char
	dbnz	R17,PC-1-1-1-1-1
	;
	call	LCDclrL1
	ldiw	R30,TCPmod*2
	call	LCDwrStrFlash
	
	call	LCDposL2
	ldiw	R30,MODname
	ldi	R17,20
	call	LCDwrStrRam
	
	;config ports/signals
	outi	ddra,0x00	;inputs
	outi	portc,0x00	;hi-z
	sbi	pROE,ROE	;SRAM OE=1
	BtInh	0
	ret	


	
	

;------------------------------------------------------------------------------------------------
;	Load MOD - load parameters
;------------------------------------------------------------------------------------------------
	;------ magic constants/routines/macros ------
DMod4Chn:	.db	4,"M.K.","M!K!","M&K!","N.T."	;4chn, 31instr.
DMod8Chn:	.db	2,"CD81","OKTA"	;8chn, 31instr.
DMod16Cn:	.db	1,"16CN"	;16chn, 31instr.
DMod32Cn:	.db	1,"32CN"	;32chn, 31instr.
	;compare m.c.'s 
MODidentCSS:	lpm	R24,Z+
	clt
MODidentCSSL1:	lpmd	R20,Z+
	cpd	R16,R20
	brne	PC+1+1+1
	set
	rjmp	PC+1+2
	dbnz	R24,MODidentCSSL1
	ret
	;compare m.c.
	.macro	mCPmc3
	cpi	@0,@2
	ldi	@1,@3
	cpc	@0H,@1
	ldi	@1,@4
	cpc	@0B3,@1
	.endm
	;compare m.c.
	.macro	mCPmc2
	cpi	@0,@2
	ldi	@1,@3
	cpc	@0H,@1
	.endm
	
	
	;====== MAIN ROUTINE ======
LoadMOD:	;config ports/signals
	outi	ddra,0xFF	;outputs
	outi	portc,0xFF	;pull-up
	cbi	pROE,ROE	;SRAM OE=0
		
	;SRAM address steping (R5:R6:R7)
	clr	R5
	inc	R5
	clrw	R6	;+1
	
	;SRAM address (R2:R3:R4)=0
	clrt	R2
	
	;------ get song title ------
	;clear string
	ldiw	R30,MODname
	ldi	R16,20
	ldi	R17,' '
LoadMODnC:	st	Z+,R17
	dbnz	R16,LoadMODnC
	;read name 
	ldiw	R30,MODname
	ldi	R17,20	
LoadMODnL:	rcall	RamGetByteA
	cpi	R16,0
	breq	LoadMODeN
	st	Z+,R16
	dbnz	R17,LoadMODnL
LoadMODeN:	
	;print song title
	ldiw	R30,TModN*2
	call	UsTxStrFlash
	ldiw	R30,MODname
	ldi	R16,20
	call	UsTxStrRam
	ldi	R16,0x0D
	call	UsTxChar
	
	call	LCDclrL1
	ldiw	R30,MODname
	ldi	R17,20
	call	LCDwrStrRam
	
	
	
	;------ test MOD format ------
	stsi	MODinsn,31
	;get magic constant
	ldiw	R16,1080
	movw	R2,R16
	clr	R4
	rcall	RamGetByteA
	push	R16
	rcall	RamGetByteA
	mov	R17,R16
	rcall	RamGetByteA
	mov	R18,R16
	rcall	RamGetByteA
	mov	R19,R16
	pop	R16
	;	
	ldiw	R30,DMod4Chn*2
	rcall	MODidentCSS
	brtc	LoadMODin4C
	;4 channel
	ldi	R30,4	;4chn
	rjmp	LoadMODid
		
LoadMODin4C:	ldiw	R30,DMod8Chn*2
	rcall	MODidentCSS
	brtc	LoadMODin8C
	;8 channel
	ldi	R30,8	;8chn
	rjmp	LoadMODid
		
LoadMODin8C:	mCPmc3	R16,R20,'F','L','T'
	breq	PC+1+5+1
	mCPmc3	R16,R20,'T','D','Z'
	brne	LoadMODinFLT
	cpi	R19,'4'
	brlo	LoadMODinFLT
	cpi	R19,'9'+1
	brsh	LoadMODinFLT
	;"FLTx", "TDZx", where x={4..9}
	mov	R30,R19
	subi	R30,'4'-4
	rjmp	LoadMODid
	
LoadMODinFLT:	mCPmc3	R17,R20,'C','H','N'
	brne	LoadMODinCHN
	cpi	R16,'4'
	brlo	LoadMODinCHN
	cpi	R16,'9'+1
	brsh	LoadMODinCHN
	;"xCHN", where x={4..9}
	mov	R30,R16
	subi	R30,'4'-4
	rjmp	LoadMODid

LoadMODinCHN:	mCPmc2	R18,R20,'C','H'
	brne	LoadMODinCH
	cpi	R16,'1'
	breq	PC+1+1+1
	cpi	R16,'2'
	brne	LoadMODin12CH
	cpi	R17,'0'
	brlo	LoadMODin12CH
	cpi	R17,'9'+1
	brsh	LoadMODin12CH
	;"xxCH", where xx={10..29}
LoadMODiy3CH:	subi	R16,'1'-1
	ldi	R18,10
	mul	R16,R18
	mov	R30,R0
	subi	R17,'0'
	add	R30,R17
	rjmp	LoadMODid
	
LoadMODin12CH:	cpi	R16,'3'
	brne	LoadMODinCH
	cpi	R17,'0'
	brlo	LoadMODinCH
	cpi	R17,'2'+1
	brlo	LoadMODiy3CH	;"xxCH", where xx={30..32}
	
LoadMODinCH:	ldiw	R30,DMod16Cn*2
	rcall	MODidentCSS
	brtc	LoadMODin16CN
	;16 channel
	ldi	R30,16	;16chn
	rjmp	LoadMODid
	
LoadMODin16CN:	ldiw	R30,DMod32Cn*2
	rcall	MODidentCSS
	brtc	LoadMODin32CN
	;32 channel
	ldi	R30,32	;32chn
	rjmp	LoadMODid
	;not identified
LoadMODin32CN:	stsi	MODinsn,15
	ldi	R30,4
LoadMODid:	sts	MODchnn,R30
	
	
	cpi	R30,MAXCH+1
	brlo	LoadMODvalf
	;invalid format
LoadMODinvf:	sts	FileLen+0,R6
	sts	FileLen+1,R6
	sts	FileLen+2,R6		
	clt
	ret		;leave if not valid fmt
LoadMODvalf:

			
	;------ read MOD patterns ------
	;create address
	lds	R16,MODinsn
	ldi	R18,20
	mov	R2,R18
	clr	R3	;sample list offset
	ldi	R17,30	;sample list item len
	mul	R16,R17
	addw	R2,R0
	
	;read sond len (patterns)
	rcall	RamGetByteA
	sts	MODpatn,R16
	
	ldi	R16,1
	add	R2,R16
	adc	R3,R4	;skip end jump item	
	
	;read pattern list
	ldiw	R30,MODpatl
	ldi	R24,128	;for each
LoadMODpL:	rcall	RamGetByteA
	st	Z+,R16
	dbnz	R24,LoadMODpL	
	
	;calc. real pattern count
	clr	R18	;lowest pattern ID
	ldiw	R30,MODpatl
	ldi	R24,128
LoadMODpL1:	ld	R16,Z+
	cp	R16,R18
	brlo	PC+1+1
	;same or higher id - store as real pattern count
	mov	R18,R16
	cpi	R16,128
	brsh	LoadMODinvf	;leave if number >= 128 
	dbnz	R24,LoadMODpL1
	inc	R18	;++ (addressing from 0)
	sts	MODpats,R18	;store it
		
	;------ read sample informations ------
	;sample list start
	ldi	R16,20
	mov	R2,R16
	clr	R3
	
	;samples data start (R11:R12:R13)
	ldiw	R16,20+31*30+1+1+128+4	;title+instr*30+some_garbage+mahoney@kaktus
	lds	R18,MODinsn
	cpi	R18,31
	breq	LoadMOD31i
	ldiw	R16,20+15*30+1+1+128	;title+instr*30+some_garbage
LoadMOD31i:	mov	R11,R16
	mov	R12,R17
	mov	R13,R4
	lds	R16,MODpats	;patterns
	lds	R17,MODchnn	;channels
	mul	R16,R17
	addw	R12,R0	;add pattern len
	
	
		
	;for each sample
	ldiw	R30,MODsmpl	;sample structure pointer
	lds	R24,MODinsn	
LoadMODsL:	ldi	R16,22
	add	R2,R16
	adc	R3,R4	;skip sample title
	;get sample len
	rcall	RamGetByteA
	mov	R9,R16
	rcall	RamGetByteA
	mov	R8,R16
	ldi	R16,2
	clr	R15
	cp	R8,R16
	cpc	R9,R4	;at least 2bytes?
	brsh	LoadMODlNS
	clrw	R8	;=0, not a sample (just some Amiga times garbage!)
	dec	R15	;invalid sample
LoadMODlNS:	clr	R10
	lslt	R8	;sample_len*=2
	addt	R8,R11	;sample_len+=sample_data_offset
	mov	R17,R11
	movw	R18,R12	;sample start address
	subit	R17,-2	;skip first 2 bytes (R17:R18:R19 - start address)
	pusht	R8	;store new sample address 
	ldi	R16,1
	add	R2,R16
	adc	R3,R4	;skip fine tune item
	;get volume
	rcall	RamGetByteA
	cpi	R16,64
	brlo	PC+1+1
	ldi	R16,64	;cut vol 0-64
	std	Z+0,R16	;store volume
	stdt	Z+1,R17	;store sample start address	
	;get repeat offset
	rcall	RamGetByteA
	mov	R20,R16
	rcall	RamGetByteA
	mov	R19,R16
	clr	R21
	lslt	R19	;sample_repeat*=2
	;get loop len
	rcall	RamGetByteA
	mov	R17,R16
	rcall	RamGetByteA
	clr	R14	;loop off (default)
	cpi	R16,2
	cpc	R17,R4
	brlo	LoadMODrn
	;loop active!
	dec	R14	;loop on
	clr	R18
	lslt	R16
	stdt	Z+7,R16	;store loop len
	addt	R16,R19	;+=repeat_offset
	addt	R16,R11	;+=sample_data_offset
	subit	R16,-1
	movw	R8,R16
	mov	R10,R18	;rewrite sample end
LoadMODrn:	popt	R11	;new sample address
	ldi	R16,1
	sub	R8,R16
	sbc	R9,R4
	sbc	R10,R4	;sample len -=1
	stdt	Z+4,R8	;store sample stop address
	std	Z+10,R14	;store loop activity (0x00/0xFF => loop off/on)
	
	tst	R15
	brne	PC+1+1	;skip if invalid sample
	rcall	MODcorSmpl	;sample data correction
	
	adiw	R30,MODsmplSize
	djnz	R24,LoadMODsL

	;config ports/signals
	outi	ddra,0x00	;inputs
	outi	portc,0x00	;hi-z
	sbi	pROE,ROE	;SRAM OE=1
	set
	ret
	
	
MODcorSmpl:	BtInh	1
	pusht	R2
	;--- loop correction ---
	;if linear interpolation is enabled it's necessary to duplicate first sample of loop data
	;to address that lies just above last loop data sample (linear interpolation always
	;needs 1 aditional sample which can lie outside of valid data)
	ldd	R16,Z+10
	cpi	R16,0xFF
	brne	MODcorSmplNL
	;loop active, try to duplicate sample
	lddt	R2,Z+4	;sample data end address
	cp	R2,R6
	cpc	R3,R6
	lds	R17,SRAMsize
	cpc	R4,R17
	brsh	MODcorSmplLOR	;skip if outside of valid data
	
	;read sample
	lddt	R18,Z+7	;loop len
	subt	R2,R18	;first sample of loop data
	rcall	RamGetByteA	;read it
	mov	R15,R16
		
	;write sample
	sbi	pROE,ROE	;OE=1
	outi	ddrc,0xFF	;ouputs
	;
	lddt	R2,Z+4	;first sample after loop data address
	mov	R16,R15
	rcall	RamPutByteA	;store sample
	;
	outi	ddrc,0x00	;inputs
	outi	portc,0xFF	;pull-up
	cbi	pROE,ROE	;SRAM OE=0
			
MODcorSmplLOR:
MODcorSmplNL:		
	;--- ramp generation ---
	.equ	MRLEN=16	;ramp len in samples
	
	ldd	R16,Z+10
	cpi	R16,0xFF
	brne	MODcorSmplR1NL
	;loop active, ramp can be generated only before loop start
	lddt	R16,Z+4
	lddt	R19,Z+7
	subt	R16,R19	;loop start
	lddt	R19,Z+1	;sample start
	subt	R19,R16	;max ramp len
	rjmp	MODcorSmplR1L
	;loop inactive, ramp can be generated only before first half of sample
MODcorSmplR1NL:	lddt	R16,Z+1	;sample start
	lddt	R19,Z+4	;sample end
	subt	R19,R16
	lsrt	R19	;max ramp len	
MODcorSmplR1L:	;
	cpith	R19,R16,MRLEN
	rjlo	MODcorSmplR1NC
	;can do start ramp
	ldiw	R16,255<<8
	ldi	R18,MRLEN
	clr	R19
	rcall	div16u
	movw	R14,R16	;volume step [fx 8.8]
	clrw	R18	;volume [fx 8.8]
	;		
	lddt	R2,Z+1
	lds	R17,SRAMsize
	ldi	R23,MRLEN
MODcorSmplR1rL:	rcall	MODcorSRamp
	addw	R18,R14	;volume step
	dbnz	R23,MODcorSmplR1rL
	;
	ldd	R16,Z+10
	cpi	R16,0xFF
	breq	MODcorSmplR1NC
	;end ramp allowed
	lddt	R2,Z+4
	ldiw	R16,MRLEN
	subw	R2,R16
	sbc	R4,R6
	;
	ldiw	R18,255<<8
	lds	R17,SRAMsize
	ldi	R23,MRLEN
MODcorSmplR2rL:	rcall	MODcorSRamp
	subw	R18,R14	;volume step
	dbnz	R23,MODcorSmplR2rL
	;
MODcorSmplR1NC:	popt	R2
	BtInh	0
	ret
	;ramp generator - signed sample volume R19={0..255}  
MODcorSRamp:	cp	R2,R6
	cpc	R3,R6
	cpc	R4,R17
	brsh	MODcorSRampS	;skip if invalid SRAM address
	;
	clr	R5
	rcall	RamGetByteA
	inc	R5
	;	
	sbi	pROE,ROE	;OE=1
	outi	ddrc,0xFF	;ouputs
	;
	mulsu	R16,R19	;apply volume
	mov	R16,R1
	rcall	RamPutByteA	;store sample
	;
	outi	ddrc,0x00	;inputs
	outi	portc,0xFF	;pull-up
	cbi	pROE,ROE	;SRAM OE=0
MODcorSRampS:	ret


	


;------------------------------------------------------------------------------------------------
;	Get byte: R16=SRAM(R2:R3:R4), R2:R3:R4+=R5:R6:R7 
;------------------------------------------------------------------------------------------------
RamGetByteA:	out	porta,R4
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	out	porta,R3
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	out	porta,R2
	.if	HCOC==1
	nop
	.endif
	addt	R2,R5	;address+=R5:R6:R7
	in	R16,pinc
	ret
	
;------------------------------------------------------------------------------------------------
;	Put byte: SRAM(R2:R3:R4)=R16, R2:R3:R4+=R5:R6:R7 
;------------------------------------------------------------------------------------------------
RamPutByteA:	out	portc,R16
	out	porta,R4
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	out	porta,R3
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	out	porta,R2
	addt	R2,R5	;address+=R5:R6:R7
	cbi	pRWR,RWR	;rw=0
	nop
	sbi	pRWR,RWR	;rw=1
	ret	
	

;------------------------------------------------------------------------------------------------
;	Receive file via USART, automatic end detection
;	- all irq's longer than few cycles should be disabled! 
;------------------------------------------------------------------------------------------------	
LoadData:	;port configuration
	call	LCDclrAll
	ldiw	R30,TRData*2
	call	LCDwrStrFlash
	
	BtInh	1
	sbi	pROE,ROE	;OE=1
	outi	ddra,0xFF	;ouputs
	outi	ddrc,0xFF	;ouputs

	;SRAM pointer
	ldit	R18,0
	
	;SRAM size
	lds	R21,SRAMsize
	
	;flush USART recieved data
	lds	R16,udr0
	ldi	R16,1
	rcall	Delay1ms
	lds	R16,udr0
	ldi	R16,1
	rcall	Delay1ms
	lds	R16,udr0
	
	
	
	ToTime	150	;1.5s timeout
	ToRes
	ToRun		;start timeout timer
	
	;wait for first byte
;LoadDataWF:	rjsc	ucsr0a,rxc0,LoadDataWF
;	rjmp	LoadDataS	;first byte in usart rx FIFO 
	
	;wait for next byte, check timeout
LoadDataWN:	lds	R16,TimOut
	cpi	R16,0xFF
	breq	LoadDataTO
	rjsc	ucsr0a,rxc0,LoadDataWN
	
	;store byte into SRAM, addr++
LoadDataS:	lds	R17,udr0	;read byte
	out	portc,R17	;send out
	out	porta,R20	;high addr
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	out	porta,R19	;med addr
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	out	porta,R18	;low addr
	subit	R18,-1	;inc address
	cbi	pRWR,RWR	;rw=0
	cp	R20,R21	;test for memory overflow
	sbi	pRWR,RWR	;rw=1
	
	brsh	LoadDataTO	;memory full - abort!	
		
	ToRes		;reset timeout
	rjmp	LoadDataWN	
	
	;timeout or full memory
LoadDataTO:	ToStop		;stop timeout timer	
	
	
	clr	R16
	tstt	R18,R16
	clt
	breq	PC+1+1+6
	set
	stst	FileLen,R18	;store file len
		
	;safe port configuration
	outi	ddra,0x00	;inputs
	outi	ddrc,0x00	;inputs
	BtInh	0
	ret	


;------------------------------------------------------------------------------------------------
;	Init MOD player data before main playing
;------------------------------------------------------------------------------------------------
InitMOD:	;reset playback structure/effects	
	ldiw	R30,MODpbck
	clr	R0
	lds	R24,MODchnn
MODipCPBCK:	;playback
	std	Z+0,R0	;stop playback
	std	Z+2+0,R0
	std	Z+2+1,R0	;0 resample coeff
	std	Z+15,R0	;no sample specified
	;pos
	std	Z+mePS+0,R0	;not specified
	;retrigger
	std	Z+meRT+0,R0	;activity
	;vibrato
	std	Z+meVI+0,R0	;activity
	std	Z+meVI+1,R0	;last parameters
	;arpeggio
	std	Z+meAR+0,R0	;activity
	;portamento
	std	Z+meTP+0,R0
	ldiw	R16,856
	stdw	Z+meTP+2,R16
	;volume slide
	std	Z+meVS+0,R0
	;
	adiw	R30,MODpbckSize
	dbnz	R24,MODipCPBCK	

	;default speed/tempo
	stsi	MODspd,6
	stsi	MODbpm,125
	stsi	MODtimc,0	
	
	;calculate tick divisor
	ldit	R16,5*SREAL	;5*sample_rate(real one!)
	lds	R19,MODbpm
	clrw	R20
	lslw	R19	;2*bpm
	rcall	div24u	;R16:R17=sample_rate/(0.4*bpm)
	stsw	MODtdiv,R16	;store it
	
	;init pattern reader vars
	stsi	MODpat,-1	;initial pattern (no pattern code)
	stsi	MODpln,64	;initial pattern line (new line code)
	stsi	MODpnln,0	;first line of new pattern (used for pattern break effect)
	ret	
	
	


;************************************************************************************************
;	External SRAM check 
;************************************************************************************************
SRAMtest:	;config ports/signals
	BtInh	1
	outi	ddra,0xFF	;outputs
	
	call	LCDclrAll
	ldiw	R30,TSRchk*2
	call	LCDwrStrFlash
	
	;------ fast mem size detection in 64kiB blocks ------
	clr	R3
	clrt	R5	;SRAM address step
	inc	R5

	ldi	R17,0x00	;initial block
	ldi	R25,128	;max count of tested blocks
	;write mark to new block {R17}
SRAMtestBTL2:	sbi	pROE,ROE	;SRAM OE=1
	outi	ddrc,0xFF	;outputs
	;
	clr	R2
	mov	R4,R17	;select block
	;
	mov	R16,R17
	rcall	RamPutByteA
	mov	R16,R17
	com	R16
	rcall	RamPutByteA
	inc	R17	;next block

	;check all marks of all written blocks {0...(R17-1)}
	outi	ddrc,0x00	;inputs
	outi	portc,0xFF	;pull-ups
	cbi	pROE,ROE	;SRAM OE=0
	;
	ldi	R18,0x00
	mov	R24,R17
SRAMtestBTL1:	clr	R2
	mov	R4,R18	;select block
	rcall	RamGetByteA
	cp	R16,R18
	brne	SRAMtestBTF	;done if not eq.
	rcall	RamGetByteA
	com	R18
	cp	R16,R18
	brne	SRAMtestBTF	;done if not eq.
	com	R18
	inc	R18
	;	
	dbnz	R24,SRAMtestBTL1
	;
	dbnz	R25,SRAMtestBTL2
	;all successfully done, R17 contains valid block count
	rjmp	PC+1+1
SRAMtestBTF:	dec	R17
	sts	SRAMsize,R17
	cpi	R17,0
	.if	MEMTST!=0
	rjeq	SRAMtestRF	;skip if no mem detected
	.else
	brne	SRAMtestRS
	.endif


	
	
	;------ full mem test of detected size ------
	.if	MEMTST!=0
	clr	R0	;0
	clr	R1
	inc	R1	;1
	;write ids ...
	sbi	pROE,ROE	;SRAM OE=1
	outi	ddrc,0xFF	;outputs
	;
	clrd	R2	
	lds	R18,SRAMsize
SRAMtestWL3:	out	porta,R4	;high addr
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	
	ldi	R17,256
SRAMtestWL2:	out	porta,R3	;med addr
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	;
	ldi	R16,256/4
	;store DWORD
SRAMtestWL1:	out	porta,R2	;low addr
	out	portc,R2	;send out
	adiw	R30,0
	cbi	pRWR,RWR	;rw=0
	nop
	sbi	pRWR,RWR	;rw=1
	inc	R2
	;
	out	porta,R2	;low addr
	out	portc,R3	;send out
	adiw	R30,0
	cbi	pRWR,RWR	;rw=0
	nop
	sbi	pRWR,RWR	;rw=1
	inc	R2
	;
	out	porta,R2	;low addr
	out	portc,R4	;send out
	adiw	R30,0
	cbi	pRWR,RWR	;rw=0
	nop
	sbi	pRWR,RWR	;rw=1
	inc	R2
	;
	out	porta,R2	;low addr
	out	portc,R5	;send out
	adiw	R30,0
	cbi	pRWR,RWR	;rw=0
	nop
	sbi	pRWR,RWR	;rw=1
	inc	R2
	;
	dbnz	R16,SRAMtestWL1
	;
	add	R3,R1
	adc	R4,R0
	adc	R5,R0
	;
	dbnz	R17,SRAMtestWL2
	;
	dbnz	R18,SRAMtestWL3
	
	;read and check ids
	outi	ddrc,0x00	;inputs
	outi	portc,0xFF	;pull-ups
	cbi	pROE,ROE	;SRAM OE=0
	;
	ldi	R24,3
	ldi	R25,4
	clrd	R2	
	lds	R18,SRAMsize
SRAMtestRL3:	out	porta,R4	;high addr
	cbi	pRHA,RHA
	sbi	pRHA,RHA
	
	ldi	R17,256
SRAMtestRL2:	out	porta,R3	;med addr
	cbi	pRMA,RMA
	sbi	pRMA,RMA
	;
	ldi	R16,256/4
	;
SRAMtestRL1:	out	porta,R2	;low addr
	inc	R2
	nop
	nop	
	.if	HCOC==1
	nop
	.endif
	in	R20,pinc
	;
	out	porta,R2	;low addr
	inc	R2
	nop
	nop	
	.if	HCOC==1
	nop
	.endif
	in	R21,pinc
	;
	out	porta,R2	;low addr
	inc	R2
	nop
	nop	
	.if	HCOC==1
	nop
	.endif
	in	R22,pinc
	;
	out	porta,R2	;low addr
	sub	R2,R24
	cp	R20,R2
	cpc	R21,R3
	.if	HCOC==1
	nop
	.endif
	in	R23,pinc
	
	cpc	R22,R4
	cpc	R23,R5
	brne	SRAMtestRF
	add	R2,R25
	;
	dbnz	R16,SRAMtestRL1
	;
	add	R3,R1
	adc	R4,R0
	adc	R5,R0
	;
	dbnz	R17,SRAMtestRL2
	;
	dbnz	R18,SRAMtestRL3
	;success
	rjmp	SRAMtestRS
	;failed
	.endif
SRAMtestRF:	ldiw	R30,TSRfail*2
	call	LCDwrStrFlash
	rjmp	PC
	
	;passed - print found size
SRAMtestRS:	lds	R16,SRAMsize
	ldi	R17,64
	mul	R16,R17
	movw	R16,R0
	clr	R18
	call	LCDwrBCD8sz
	ldiw	R30,TUkB*2
	call	LCDwrStrFlash
	ldi	R16,70
	rcall	Delay10ms

	;safe port configuration
	sbi	pROE,ROE	;SRAM OE=1
	outi	ddra,0x00	;inputs
	outi	ddrc,0x00	;inputs
	
	BtInh	0
	ret	
	

;************************************************************************************************
;	Delay routines
;************************************************************************************************
;------------------------------------------------------------------------------------------------
;	R16*1ms delay (0=~256ms)
;------------------------------------------------------------------------------------------------
Delay1ms:	;~1ms
	ldi	R17,XTAL/((100*3)*1000)
	;100*3 cycles
Delay1msL2:	ldi	R18,100
Delay1msL1:	dbnz	R18,Delay1msL1
	dbnz	R17,Delay1msL2
	dbnz	R16,Delay1ms	
	ret
	
;------------------------------------------------------------------------------------------------
;	R16*10ms delay (0=~2.56s)
;------------------------------------------------------------------------------------------------
Delay10ms:	mov	R15,R16
	ldi	R16,10
	rcall	Delay1ms	;10ms delay
	dbnz	R15,PC-1-1-1
	ret	
	
	
;------------------------------------------------------------------------------------------
; div16u - 16/16 bit unsigned division, max. speed optimized
; deleni dvou neznaminkovych 16-bit cisel, maximalni optimalizace na rychlost
;
; dDd16r0u:dDd16r1u <- dDd16r0u:dDd16r1u div dDs16r0u:dDs16r1u
; dRe16r0u:dRe16r1u <- dDd16r0u:dDd16r1u mod dDs16r0u:dDs16r1u
;
; words:  142 + ret                        (puvodni od ATMELu: 196 + ret)
; cycles: 126/134/142 + ret [min/avg/max]  (puvodni od ATMELu: 148/172/196 + ret)
;
; Author: Pavel Borovsky, profesor@ti.cz, Czech Republic, Europe 3.6.2000
;------------------------------------------------------------------------------------------
	.def	dRe16r0u=R14
	.def	dRe16r1u=R15
	.def	dDd16r0u=R16
	.def	dDd16r1u=R17
	.def	dDs16r0u=R18
	.def	dDs16r1u=R19
	;
div16u:	clr	dRe16r0u
	sub	dRe16r1u,dRe16r1u
	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_1
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_1:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_2
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_2:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_3
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_3:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_4
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_4:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_5
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_5:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_6
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_6:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_7
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_7:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_8
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_8:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_9
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_9:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_10
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_10:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_11
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_11:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_12
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_12:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_13
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_13:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_14
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_14:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_15
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_15:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_16
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_16:	rol	dDd16r0u
	rol	dDd16r1u
	com	dDd16r0u
	com	dDd16r1u
	ret

	
;------------------------------------------------------------------------------------------
; div24u - 24/24 bit unsigned division, max. speed optimized
; deleni dvou neznaminkovych 24-bit cisel, maximalne optimalizovano na rychlost
;
; dDd24r0u:dDd24r1u:dDd24r2u <- dDd24r0u:dDd24r1u:dDd24r2u div dDs24r0u:dDs24r1u:dDs24r2u
; dRe24r0u:dRe24r1u:dRe24r2u <- dDd24r0u:dDd24r1u:dDd24r2u mod dDs24r0u:dDs24r1u:dDs24r2u
;
; words:   297 + ret
; cycles:  249/273/297 + ret [min/avg/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------
	.def	dRe24r0u=R13
	.def	dRe24r1u=R14
	.def	dRe24r2u=R15
	.def	dDd24r0u=R16
	.def	dDd24r1u=R17
	.def	dDd24r2u=R18
	.def	dDs24r0u=R19
	.def	dDs24r1u=R20
	.def	dDs24r2u=R21
	;
div24u:	clr	dRe24r0u
	clr	dRe24r1u
	sub	dRe24r2u,dRe24r2u
	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_1
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_1:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_2
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_2:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_3
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_3:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_4
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_4:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_5
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_5:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_6
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_6:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_7
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_7:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_8
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_8:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_9
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_9:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_10
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_10:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_11
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_11:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_12
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_12:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_13
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_13:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_14
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_14:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_15
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_15:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_16
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_16:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_17
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_17:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_18
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_18:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_19
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_19:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_20
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_20:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_21
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_21:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_22
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_22:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_23
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_23:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_24
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_24:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	com	dDd24r0u
	com	dDd24r1u
	com	dDd24r2u
	ret
	

;------------------------------------------------------------------------------------------------------------------------
; bin2BCD8u2 - 8-bit binary to 2-digit BCD conversion (0..99)
; prevod 8-bit binarniho cisla na 2-ciferne cislo v BCD
;
; BCD16res0:BCD16res1 <- bin16r0
; BCD(0x01 0x02) <- bin(12)
;      ^
;  BCD16res1 
;
; words:  5 + ret
; cycles: 5/41 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD8res0=R16	; !!! BCD8res0 = bin8r0 !!!
	.def	BCD8res1=R17
	.def	bin8r0=R16
	;
bin2BCD8u2:	ser	BCD8res1
BB8u2_L10:	inc	BCD8res1
	subi	bin8r0,10
	brcc	BB8u2_L10
	subi	bin8r0,-10
	ret
	
;------------------------------------------------------------------------------------------------------------------------
; bin2BCD8u3 - 8-bit binary to 3-digit BCD conversion (0..255)
; prevod 8-bit binarniho cisla na 3-ciferne cislo v BCD
;
; BCD16res0:BCD16res1:BCD16res2 <- bin16r0
; BCD(0x01 0x02 0x03) <- bin(123)
;      ^
;  BCD16res2
;
; words:  10 + ret
; cycles: 10/50 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD8res0=R16	; !!! BCD8res0 = bin8r0 !!!
	.def	BCD8res1=R17
	.def	BCD8res2=R18
	.def	bin8r0=R16
	;
bin2BCD8u3:	ser	BCD8res2
	ser	BCD8res1
BB8u3_L100:	inc	BCD8res2
	subi	bin8r0,100
	brcc	BB8u3_L100
	subi	bin8r0,-100
BB8u3_L10:	inc	BCD8res1
	subi	bin8r0,10
	brcc	BB8u3_L10
	subi	bin8r0,-10
	ret
	
;------------------------------------------------------------------------------------------------------------------------
; bin2BCD24p8 - 24-bit binary to 8-digit packed BCD conversion (0..16777215)
; prevod 24-bit binarniho cisla na osmiciferne cislo v BCD
;
; BCD24res0:BCD24res1:BCD24res2:BCD24res3 <- bin24r0:bin24r1:bin24r2
; BCD(0x12 0x34 0x56 0x78) <- bin(12345678)
;      ^              ^
; BCD16res3      BCD16res0
;
; words:  59 + ret
; cycles: 59/347  + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD24res0=R12
	.def	BCD24res1=R13
	.def	BCD24res2=R14
	.def	BCD24res3=R15
	.def	bin24r0=R16
	.def	bin24r1=R17
	.def	bin24r2=R18
	.def	BB24_cnt1=R30
	.def	BB24_cnt2=R31

bin2BCD24p8:	ser	BB24_cnt2
	ldi	BB24_cnt1,(256-16)
BB24p8_L10M:	subi	BB24_cnt1,(-16)
	subi	bin24r0,0x80
	sbci	bin24r1,0x96
	sbci	bin24r2,0x98	; -10000000
	brcc	BB24p8_L10M
	subi	bin24r0,0x80
	sbci	bin24r1,0x69
	sbci	bin24r2,0x67	; +10000000
	mov	BCD24res3,BB24_cnt2
BB24p8_L1M:	inc	BCD24res3
	subi	bin24r0,0x40
	sbci	bin24r1,0x42
	sbci	bin24r2,0x0F	; -1000000
	brcc	BB24p8_L1M
	subi	bin24r0,0xC0
	sbci	bin24r1,0xBD
	sbci	bin24r2,0xF0	; +1000000
	or	BCD24res3,BB24_cnt1
	ldi	BB24_cnt1,(256-16)
BB24p8_L100k:	subi	BB24_cnt1,(-16)
	subi	bin24r0,0xA0
	sbci	bin24r1,0x86
	sbci	bin24r2,0x01	; -100000
	brcc	BB24p8_L100k
	subi	bin24r0,0x60
	sbci	bin24r1,0x79
	sbci	bin24r2,0xFE	; +100000
	mov	BCD24res2,BB24_cnt2
BB24p8_L10k:	inc	BCD24res2
	subi	bin24r0,0x10
	sbci	bin24r1,0x27
	sbci	bin24r2,0x00	; -10000
	brcc	BB24p8_L10k
	subi	bin24r0,0xF0
	sbci	bin24r1,0xD8	; +10000
	or	BCD24res2,BB24_cnt1
	ldi	BB24_cnt1,(256-16)
BB24p8_L1k:	subi	BB24_cnt1,(-16)
	subi	bin24r0,low(1000)
	sbci	bin24r1,high(1000)	; -1000
	brcc	BB24p8_L1k
	subi	bin24r0,low(-1000)
	sbci	bin24r1,high(-1000)	; +1000
	mov	BCD24res1,BB24_cnt2
BB24p8_L100:	inc	BCD24res1
	subi	bin24r0,low(100)
	sbci	bin24r1,high(100)	; -100
	brcc	BB24p8_L100
	subi	bin24r0,low(-100)	; +100
	or	BCD24res1,BB24_cnt1
	ldi	BB24_cnt1,(256-16)
BB24p8_L10:	subi	BB24_cnt1,(-16)
	subi	bin24r0,10	; -10
	brcc	BB24p8_L10
	subi	bin24r0,-10	; +10
	mov	BCD24res0,bin24r0
	or	BCD24res0,BB24_cnt1
	ret
	
	
	.include "MOD_tables.asm"
	.include "text.asm"

;################################################################################################
;	THIS MUST BE AT VERY END OF MAIN MODULE AFTER ALL INCLUDES
;################################################################################################
	.dseg
REND:	.byte	0
	.if	REND>=RAMEND-STSIZE
	.error	"RAM IS FULL!!!"
	.endif
	.cseg
	